"""
All the ipyvuetify class override as SepalWidgets.

This file overwrite all the widgets generating by ipyvuetify to add the SepalWidget class as a parent.
It should not be modified from here as it's automatically generated.

last update: 2022-12-29
"""

import ipyvuetify as v

from sepal_ui.sepalwidgets.sepalwidget import SepalWidget


class Html(v.Html, SepalWidget):
    pass


class Alert(v.Alert, SepalWidget):
    pass


class App(v.App, SepalWidget):
    pass


class AppBar(v.AppBar, SepalWidget):
    pass


class AppBarNavIcon(v.AppBarNavIcon, SepalWidget):
    pass


class Autocomplete(v.Autocomplete, SepalWidget):
    pass


class Avatar(v.Avatar, SepalWidget):
    pass


class Badge(v.Badge, SepalWidget):
    pass


class Banner(v.Banner, SepalWidget):
    pass


class BottomNavigation(v.BottomNavigation, SepalWidget):
    pass


class BottomSheet(v.BottomSheet, SepalWidget):
    pass


class Breadcrumbs(v.Breadcrumbs, SepalWidget):
    pass


class BreadcrumbsDivider(v.BreadcrumbsDivider, SepalWidget):
    pass


class BreadcrumbsItem(v.BreadcrumbsItem, SepalWidget):
    pass


class Btn(v.Btn, SepalWidget):
    pass


class BtnToggle(v.BtnToggle, SepalWidget):
    pass


class Calendar(v.Calendar, SepalWidget):
    pass


class CalendarDaily(v.CalendarDaily, SepalWidget):
    pass


class CalendarMonthly(v.CalendarMonthly, SepalWidget):
    pass


class CalendarWeekly(v.CalendarWeekly, SepalWidget):
    pass


class Card(v.Card, SepalWidget):
    pass


class CardActions(v.CardActions, SepalWidget):
    pass


class CardSubtitle(v.CardSubtitle, SepalWidget):
    pass


class CardText(v.CardText, SepalWidget):
    pass


class CardTitle(v.CardTitle, SepalWidget):
    pass


class Carousel(v.Carousel, SepalWidget):
    pass


class CarouselItem(v.CarouselItem, SepalWidget):
    pass


class CarouselReverseTransition(v.CarouselReverseTransition, SepalWidget):
    pass


class CarouselTransition(v.CarouselTransition, SepalWidget):
    pass


class Checkbox(v.Checkbox, SepalWidget):
    pass


class Chip(v.Chip, SepalWidget):
    pass


class ChipGroup(v.ChipGroup, SepalWidget):
    pass


class Col(v.Col, SepalWidget):
    pass


class ColorPicker(v.ColorPicker, SepalWidget):
    pass


class ColorPickerCanvas(v.ColorPickerCanvas, SepalWidget):
    pass


class ColorPickerSwatches(v.ColorPickerSwatches, SepalWidget):
    pass


class Combobox(v.Combobox, SepalWidget):
    pass


class Container(v.Container, SepalWidget):
    pass


class Content(v.Content, SepalWidget):
    pass


class Counter(v.Counter, SepalWidget):
    pass


class Data(v.Data, SepalWidget):
    pass


class DataFooter(v.DataFooter, SepalWidget):
    pass


class DataIterator(v.DataIterator, SepalWidget):
    pass


class DataTable(v.DataTable, SepalWidget):
    pass


class DataTableHeader(v.DataTableHeader, SepalWidget):
    pass


class DatePicker(v.DatePicker, SepalWidget):
    pass


class DatePickerDateTable(v.DatePickerDateTable, SepalWidget):
    pass


class DatePickerHeader(v.DatePickerHeader, SepalWidget):
    pass


class DatePickerMonthTable(v.DatePickerMonthTable, SepalWidget):
    pass


class DatePickerTitle(v.DatePickerTitle, SepalWidget):
    pass


class DatePickerYears(v.DatePickerYears, SepalWidget):
    pass


class Dialog(v.Dialog, SepalWidget):
    pass


class DialogBottomTransition(v.DialogBottomTransition, SepalWidget):
    pass


class DialogTransition(v.DialogTransition, SepalWidget):
    pass


class Divider(v.Divider, SepalWidget):
    pass


class EditDialog(v.EditDialog, SepalWidget):
    pass


class ExpandTransition(v.ExpandTransition, SepalWidget):
    pass


class ExpandXTransition(v.ExpandXTransition, SepalWidget):
    pass


class ExpansionPanel(v.ExpansionPanel, SepalWidget):
    pass


class ExpansionPanelContent(v.ExpansionPanelContent, SepalWidget):
    pass


class ExpansionPanelHeader(v.ExpansionPanelHeader, SepalWidget):
    pass


class ExpansionPanels(v.ExpansionPanels, SepalWidget):
    pass


class FabTransition(v.FabTransition, SepalWidget):
    pass


class FadeTransition(v.FadeTransition, SepalWidget):
    pass


class FileInput(v.FileInput, SepalWidget):
    pass


class Flex(v.Flex, SepalWidget):
    pass


class Footer(v.Footer, SepalWidget):
    pass


class Form(v.Form, SepalWidget):
    pass


class Hover(v.Hover, SepalWidget):
    pass


class Icon(v.Icon, SepalWidget):
    pass


class Img(v.Img, SepalWidget):
    pass


class Input(v.Input, SepalWidget):
    pass


class Item(v.Item, SepalWidget):
    pass


class ItemGroup(v.ItemGroup, SepalWidget):
    pass


class Label(v.Label, SepalWidget):
    pass


class Layout(v.Layout, SepalWidget):
    pass


class Lazy(v.Lazy, SepalWidget):
    pass


class List(v.List, SepalWidget):
    pass


class ListGroup(v.ListGroup, SepalWidget):
    pass


class ListItem(v.ListItem, SepalWidget):
    pass


class ListItemAction(v.ListItemAction, SepalWidget):
    pass


class ListItemActionText(v.ListItemActionText, SepalWidget):
    pass


class ListItemAvatar(v.ListItemAvatar, SepalWidget):
    pass


class ListItemContent(v.ListItemContent, SepalWidget):
    pass


class ListItemGroup(v.ListItemGroup, SepalWidget):
    pass


class ListItemIcon(v.ListItemIcon, SepalWidget):
    pass


class ListItemSubtitle(v.ListItemSubtitle, SepalWidget):
    pass


class ListItemTitle(v.ListItemTitle, SepalWidget):
    pass


class Menu(v.Menu, SepalWidget):
    pass


class MenuTransition(v.MenuTransition, SepalWidget):
    pass


class Messages(v.Messages, SepalWidget):
    pass


class NavigationDrawer(v.NavigationDrawer, SepalWidget):
    pass


class OverflowBtn(v.OverflowBtn, SepalWidget):
    pass


class Overlay(v.Overlay, SepalWidget):
    pass


class Pagination(v.Pagination, SepalWidget):
    pass


class Parallax(v.Parallax, SepalWidget):
    pass


class Picker(v.Picker, SepalWidget):
    pass


class ProgressCircular(v.ProgressCircular, SepalWidget):
    pass


class ProgressLinear(v.ProgressLinear, SepalWidget):
    pass


class Radio(v.Radio, SepalWidget):
    pass


class RadioGroup(v.RadioGroup, SepalWidget):
    pass


class RangeSlider(v.RangeSlider, SepalWidget):
    pass


class Rating(v.Rating, SepalWidget):
    pass


class Responsive(v.Responsive, SepalWidget):
    pass


class Row(v.Row, SepalWidget):
    pass


class ScaleTransition(v.ScaleTransition, SepalWidget):
    pass


class ScrollXReverseTransition(v.ScrollXReverseTransition, SepalWidget):
    pass


class ScrollXTransition(v.ScrollXTransition, SepalWidget):
    pass


class ScrollYReverseTransition(v.ScrollYReverseTransition, SepalWidget):
    pass


class ScrollYTransition(v.ScrollYTransition, SepalWidget):
    pass


class Select(v.Select, SepalWidget):
    pass


class Sheet(v.Sheet, SepalWidget):
    pass


class SimpleCheckbox(v.SimpleCheckbox, SepalWidget):
    pass


class SimpleTable(v.SimpleTable, SepalWidget):
    pass


class SkeletonLoader(v.SkeletonLoader, SepalWidget):
    pass


class SlideGroup(v.SlideGroup, SepalWidget):
    pass


class SlideItem(v.SlideItem, SepalWidget):
    pass


class SlideXReverseTransition(v.SlideXReverseTransition, SepalWidget):
    pass


class SlideXTransition(v.SlideXTransition, SepalWidget):
    pass


class SlideYReverseTransition(v.SlideYReverseTransition, SepalWidget):
    pass


class SlideYTransition(v.SlideYTransition, SepalWidget):
    pass


class Slider(v.Slider, SepalWidget):
    pass


class Snackbar(v.Snackbar, SepalWidget):
    pass


class Spacer(v.Spacer, SepalWidget):
    pass


class Sparkline(v.Sparkline, SepalWidget):
    pass


class SpeedDial(v.SpeedDial, SepalWidget):
    pass


class Stepper(v.Stepper, SepalWidget):
    pass


class StepperContent(v.StepperContent, SepalWidget):
    pass


class StepperHeader(v.StepperHeader, SepalWidget):
    pass


class StepperItems(v.StepperItems, SepalWidget):
    pass


class StepperStep(v.StepperStep, SepalWidget):
    pass


class Subheader(v.Subheader, SepalWidget):
    pass


class Switch(v.Switch, SepalWidget):
    pass


class SystemBar(v.SystemBar, SepalWidget):
    pass


class Tab(v.Tab, SepalWidget):
    pass


class TabItem(v.TabItem, SepalWidget):
    pass


class TabReverseTransition(v.TabReverseTransition, SepalWidget):
    pass


class TabTransition(v.TabTransition, SepalWidget):
    pass


class TableOverflow(v.TableOverflow, SepalWidget):
    pass


class Tabs(v.Tabs, SepalWidget):
    pass


class TabsItems(v.TabsItems, SepalWidget):
    pass


class TabsSlider(v.TabsSlider, SepalWidget):
    pass


class Text(v.Text, SepalWidget):
    pass


class TextField(v.TextField, SepalWidget):
    pass


class Textarea(v.Textarea, SepalWidget):
    pass


class ThemeProvider(v.ThemeProvider, SepalWidget):
    pass


class TimePicker(v.TimePicker, SepalWidget):
    pass


class TimePickerClock(v.TimePickerClock, SepalWidget):
    pass


class TimePickerTitle(v.TimePickerTitle, SepalWidget):
    pass


class Timeline(v.Timeline, SepalWidget):
    pass


class TimelineItem(v.TimelineItem, SepalWidget):
    pass


class Toolbar(v.Toolbar, SepalWidget):
    pass


class ToolbarItems(v.ToolbarItems, SepalWidget):
    pass


class ToolbarTitle(v.ToolbarTitle, SepalWidget):
    pass


class Tooltip(v.Tooltip, SepalWidget):
    pass


class Treeview(v.Treeview, SepalWidget):
    pass


class TreeviewNode(v.TreeviewNode, SepalWidget):
    pass


class VirtualTable(v.VirtualTable, SepalWidget):
    pass


class Window(v.Window, SepalWidget):
    pass


class WindowItem(v.WindowItem, SepalWidget):
    pass



"""
Customized ``Control`` to display the value of all available layers on a specific pixel.
"""

import json
from pathlib import Path
from typing import Optional, Sequence, Union

import ee
import geopandas as gpd
import ipyvuetify as v
import rasterio as rio
import rioxarray
from ipyleaflet import GeoJSON, Map
from rasterio.crs import CRS
from shapely import geometry as sg

from sepal_ui import color
from sepal_ui import sepalwidgets as sw
from sepal_ui.frontend import styles as ss
from sepal_ui.mapping.layer import EELayer
from sepal_ui.mapping.menu_control import MenuControl
from sepal_ui.message import ms
from sepal_ui.scripts import decorator as sd


class ValueInspector(MenuControl):

    m: Optional[Map] = None
    "the map on which he vinspector is displayed to interact with it's layers"

    w_loading: Optional[v.ProgressLinear] = None
    "The progress bar on top of the Card"

    menu: Optional[v.Menu] = None
    "The menu displayed when the map btn is clicked"

    text: Optional[v.CardText] = None
    "The text element from the card that is edited when the user click on the map"

    def __init__(self, m: Map, **kwargs) -> None:
        """
        Widget control displaying a btn on the map.

        When clicked the menu expand to show the values of each layer available on the map. The menu values will be change when the user click on a location on the map. It can digest any Layer added on a SepalMap.

        Args:
            m: the map on which he vinspector is displayed to interact with it's layers
        """
        # set some default parameters
        kwargs.setdefault("position", "topleft")
        kwargs["m"] = m

        # create a loading to place it on top of the card. It will always be visible
        # even when the card is scrolled
        p_style = json.loads((ss.JSON_DIR / "progress_bar.json").read_text())
        self.w_loading = sw.ProgressLinear(
            indeterminate=False,
            background_color=color.menu,
            color=p_style["color"][v.theme.dark],
        )

        # set up the content
        title = sw.CardTitle(children=[ms.v_inspector.title])
        self.text = sw.CardText(children=[ms.v_inspector.landing])

        # create the menu widget
        super().__init__("fa-solid fa-crosshairs", self.text, title, **kwargs)

        # adapt the size
        self.set_size(min_height=0)

        # add js behaviour
        self.menu.observe(self.toggle_cursor, "v_model")
        self.m.on_interaction(self.read_data)

    def toggle_cursor(self, *args) -> None:
        """
        Toggle the cursor display.

        Toggle the cursor on the map to notify to the user that the inspector
        mode is activated.
        """
        cursors = [{"cursor": "grab"}, {"cursor": "crosshair"}]
        self.m.default_style = cursors[self.menu.v_model]

        return

    def read_data(self, **kwargs) -> None:
        """
        Read the data when the map is clicked with the vinspector activated.

        Args:
            kwargs: any arguments from the map interaction
        """
        # check if the v_inspector is active
        is_click = kwargs.get("type") == "click"
        is_active = self.menu.v_model is True
        if not (is_click and is_active):
            return

        # set the loading mode. Cannot be done as a decorator to avoid
        # flickering while moving the cursor on the map
        self.w_loading.indeterminate = True
        self.m.default_style = {"cursor": "wait"}

        # init the text children
        children = []

        # get the coordinates as (x, y)
        lng, lat = coords = [c for c in reversed(kwargs.get("coordinates"))]

        # write the coordinates and the scale
        txt = ms.v_inspector.coords.format(round(self.m.get_scale()))
        children.append(sw.Html(tag="h4", children=[txt]))
        children.append(sw.Html(tag="p", children=[f"[{lng:.3f}, {lat:.3f}]"]))

        # write the layers data
        children.append(sw.Html(tag="h4", children=["Layers"]))
        layers = [lyr for lyr in self.m.layers if not lyr.base]
        for lyr in layers:
            children.append(sw.Html(tag="h5", children=[lyr.name]))

            if isinstance(lyr, EELayer):
                data = self._from_eelayer(lyr.ee_object, coords)
            elif isinstance(lyr, GeoJSON):
                data = self._from_geojson(lyr.data, coords)
            elif type(lyr).__name__ == "BoundTileLayer":
                data = self._from_raster(lyr.raster, coords)
            else:
                data = {ms.v_inspector.info.header: ms.v_inspector.info.text}

            for k, val in data.items():
                children.append(sw.Html(tag="span", children=[f"{k}: {val}"]))
                children.append(sw.Html(tag="br", children=[]))

        # set them in the card
        self.text.children = children

        # set back the cursor to crosshair
        self.w_loading.indeterminate = False
        self.m.default_style = {"cursor": "crosshair"}

        # one last flicker to replace the menu next to the btn
        # if not it goes below the map
        # i've try playing with the styles but it didn't worked out well
        # lost hours on this issue : 1h
        self.menu.v_model = False
        self.menu.v_model = True

        return

    @sd.need_ee
    def _from_eelayer(self, ee_obj: ee.ComputedObject, coords: Sequence[float]) -> dict:
        """
        extract the values of the ee_object for the considered point.

        Args:
            ee_obj: the ee object to reduce to a single point
            coords: the coordinates of the point (lng, lat).

        Returns:
            tke value associated to the image/feature names
        """
        # create a gee point
        ee_point = ee.Geometry.Point(*coords)

        if isinstance(ee_obj, ee.FeatureCollection):

            # filter all the value to the point
            features = ee_obj.filterBounds(ee_point)

            # if there is none, print non for every property
            if features.size().getInfo() == 0:
                cols = ee_obj.first().propertyNames().getInfo()
                pixel_values = {c: None for c in cols if c not in ["system:index"]}

            # else simply return all the values of the first element
            else:
                pixel_values = features.first().toDictionary().getInfo()

        elif isinstance(ee_obj, ee.Image):

            # reduce the layer region using mean
            pixel_values = ee_obj.reduceRegion(
                geometry=ee_point,
                scale=self.m.get_scale(),
                reducer=ee.Reducer.mean(),
            ).getInfo()

        else:
            raise ValueError(
                f'the layer object is a "{type(ee_obj)}" which is not accepted.'
            )

        return pixel_values

    def _from_geojson(self, data: dict, coords: Sequence[float]) -> dict:
        """
        extract the values of the data for the considered point.

        Args:
            data: the shape to reduce to a single point
            coords: the coordinates of the point (lng, lat).

        Returns:
            The value associated to the feature names
        """
        # extract the coordinates as a poin
        point = sg.Point(*coords)

        # filter the data to 1 point
        gdf = gpd.GeoDataFrame.from_features(data)
        gdf_filtered = gdf[gdf.contains(point)]
        skip_cols = ["geometry", "style"]

        # only display the columns name if empty
        if len(gdf_filtered) == 0:
            cols = gdf.columns.to_list()
            return {c: None for c in cols if c not in skip_cols}

        # else print the values of the first element
        else:
            return gdf_filtered.iloc[0, ~gdf.columns.isin(skip_cols)].to_dict()

    def _from_raster(self, raster: Union[str, Path], coords: Sequence[float]) -> dict:
        """
        extract the values of the data-array for the considered point.

        Args:
            raster: the path to the image to reduce to a single point
            coords: the coordinates of the point (lng, lat).

        Returns:
            The value associated to the feature names
        """
        # extract the coordinates as a point
        point = sg.Point(*coords)

        # extract the pixel size in degrees (equatorial appoximation)
        scale = self.m.get_scale() * 0.00001

        # open the image and unproject it
        da = rioxarray.open_rasterio(raster, masked=True)
        da = da.chunk((1000, 1000))
        if da.rio.crs != CRS.from_string("EPSG:4326"):
            da = da.rio.reproject("EPSG:4326")

        # sample is not available for da so I do as in GEE a mean reducer around 1px
        # is it an overkill ? yes
        if sg.box(*da.rio.bounds()).contains(point):
            bounds = point.buffer(scale).bounds
            window = rio.windows.from_bounds(*bounds, transform=da.rio.transform())
            da_filtered = da.rio.isel_window(window)
            means = da_filtered.mean(axis=(1, 2)).to_numpy()
            pixel_values = {
                ms.v_inspector.band.format(i + 1): v for i, v in enumerate(means)
            }

        # if the point is out of the image display None
        else:
            pixel_values = {
                ms.v_inspector.band.format(i + 1): None for i in range(da.rio.count)
            }

        return pixel_values



"""
Custom Buttons.

Gather the customized ``ipyvuetifyWidgets`` used to create buttons.
All the content of this modules is included in the parent ``sepal_ui.sepalwidgets`` package. So it can be imported directly from there.

Example:
    .. jupyter-execute::
    
        from sepal_ui import sepalwidgets as sw
        
        sw.Btn()
"""

import warnings
from pathlib import Path
from typing import Optional, Union

import ipyvuetify as v
import traitlets as t
from deprecated.sphinx import deprecated
from traitlets import observe
from typing_extensions import Self

from sepal_ui.scripts import utils as su
from sepal_ui.sepalwidgets.sepalwidget import SepalWidget

__all__ = ["Btn", "DownloadBtn"]


class Btn(v.Btn, SepalWidget):

    v_icon: Optional[v.Icon] = None
    "the icon in the btn"

    gliph: t.Unicode = t.Unicode("").tag(sync=True)
    "the name of the icon"

    msg: t.Unicode = t.Unicode("").tag(sync=True)
    "the text of the btn"

    def __init__(self, msg: str = "", gliph: str = "", **kwargs) -> None:
        """
        Custom process Btn filled with the provided text.

        The color will be defaulted to 'primary' and can be changed afterward according to your need.

        Args:
            msg: the text to display in the btn
            gliph: the full name of any mdi/fa icon
            text: the text to display in the btn
            icon: the full name of any mdi/fa icon
            kwargs (dict, optional): any parameters from v.Btn. if set, 'children' will be overwritten.

        .. deprecated:: 2.13
            ``text`` and ``icon`` will be replaced by ``msg`` and ``gliph`` to avoid duplicating ipyvuetify trait.

        .. deprecated:: 2.14
            Btn is not using a default ``msg`` anymor`.
        """
        # deprecation in 2.13 of text and icon
        # as they already exist in the ipyvuetify Btn traits (as booleans)
        if "text" in kwargs:
            if isinstance(kwargs["text"], str):
                msg = kwargs.pop("text")
                warnings.warn(
                    '"text" is deprecated, please use "msg" instead', DeprecationWarning
                )
        if "icon" in kwargs:
            if isinstance(kwargs["icon"], str):
                gliph = kwargs.pop("icon")
                warnings.warn(
                    '"icon" is deprecated, please use "gliph" instead',
                    DeprecationWarning,
                )

        # create the default v_icon
        self.v_icon = v.Icon(children=[""])

        # set the default parameters
        kwargs.setdefault("color", "primary")
        kwargs["children"] = [self.v_icon, self.msg]

        # call the constructor
        super().__init__(**kwargs)

        self.gliph = gliph
        self.msg = msg

    @observe("gliph")
    def _set_gliph(self, change: dict) -> Self:
        """
        Set a new icon. If the icon is set to "", then it's hidden.
        """
        new_gliph = change["new"]
        self.v_icon.children = [new_gliph]

        # hide the component to avoid the right padding
        if not new_gliph:
            su.hide_component(self.v_icon)
        else:
            su.show_component(self.v_icon)

        return self

    @observe("msg")
    def _set_text(self, change: dict) -> Self:
        """
        Set the text of the btn.
        """
        self.v_icon.left = bool(change["new"])
        self.children = [self.v_icon, change["new"]]

        return self

    @deprecated(version="2.14", reason="Replace by the private _set_gliph")
    def set_icon(self, icon: str = "") -> Self:
        """
        Set a new icon. If the icon is set to "", then it's hidden.

        Args:
            icon: the full name of a mdi/fa icon
        """
        self.gliph = icon
        return self

    def toggle_loading(self) -> Self:
        """
        Jump between two states : disabled and loading - enabled and not loading.
        """
        self.loading = not self.loading
        self.disabled = self.loading

        return self


class DownloadBtn(v.Btn, SepalWidget):
    def __init__(self, text: str, path: Union[str, Path] = "#", **kwargs) -> None:
        """
        Custom download Btn filled with the provided text.

        The download icon is automatically embeded and green. The btn only accepts absolute links, if non is provided then the btn stays disabled.

        Args:
            text: the message inside the btn
            path: the absoluteor relative path to a downloadable content
            kwargs: any parameter from a v.Btn. if set, 'children' and 'target' will be overwritten.
        """
        # create a download icon
        v_icon = v.Icon(left=True, children=["fa-solid fa-download"])

        # set default parameters
        kwargs.setdefault("class_", "ma-2")
        kwargs.setdefault("xs5", True)
        kwargs.setdefault("color", "success")
        kwargs["children"] = [v_icon, text]
        kwargs["target"] = "_blank"
        kwargs["attributes"] = {"download": None}

        # call the constructor
        super().__init__(**kwargs)

        # create the URL
        self.set_url(path)

    def set_url(self, path: Union[str, Path] = "#") -> Self:
        """
        Set the URL of the download btn. and unable it.

        If nothing is provided the btn is disabled.

        Args:
            path: the absolute path to a downloadable content
        """
        # set the url
        url = su.create_download_link(path)
        self.href = url

        # unable or disable the btn
        self.disabled = str(path) == "#"

        # set the download attribute
        name = None if str(path) == "#" else Path(path).name
        self.attributes = {"download": name}

        return self



"""
All modules used to create an Reclassification interface.

Package to access all the widget, model and tiles to create an Reclassifying interface. This is compatible with any type of geographic data (vector raster, ee ...).

Every module content can be called directly from the package.

Example:
    .. jupyter-execute::

        from sepal_ui import reclassify

        reclassify.ReclassifyTile(gee=False)
"""

from .parameters import *
from .reclassify_model import *
from .reclassify_tile import *
from .reclassify_view import *
from .table_view import *



"""
Customized ``Layer`` object containing EE metadata.
"""

from typing import Optional

import ee
from ipyleaflet import TileLayer


class EELayer(TileLayer):

    ee_object: Optional[ee.ComputedObject] = None
    "ee.object: the ee.object displayed on the map"

    def __init__(self, ee_object: ee.ComputedObject, **kwargs) -> None:
        """
        Wrapper of the TileLayer class to add the ee object as a member.

        useful to get back the values for specific points in a v_inspector.

        Args:
            ee_object (ee.object): the ee.object displayed on the map
        """
        self.ee_object = ee_object

        super().__init__(**kwargs)



"""
All the widgets available in sepal-ui.

Package to access all the widget available in sepal-ui. The widgets are all derived from ``IpyvuetifyWidget`` and ``SepalWidget``. They can be used to build interfaces in applications. ``sepal_ui.sepalwidgets`` include all the widgets from `ìpyvuetify`` and some extra one that are usefull to build GIS related applications.

Everything module content can be called directly from the package.

Example:
    .. jupyter-execute::

        from sepal_ui import sepalwidgets as sw

        sw.Btn()
"""

from sepal_ui.sepalwidgets.sepal_ipyvuetify import *  # noqa: I

# import and/or overwrite with our customized widgets
from sepal_ui.sepalwidgets.sepalwidget import *
from sepal_ui.sepalwidgets.alert import *
from sepal_ui.sepalwidgets.app import *
from sepal_ui.sepalwidgets.btn import *
from sepal_ui.sepalwidgets.inputs import *
from sepal_ui.sepalwidgets.tile import *
from sepal_ui.sepalwidgets.widget import *
from sepal_ui.sepalwidgets.radio import *



"""
All modules used to create AOI selection interface.

Package to access all the widget, model and tiles to create an AOI selection interface available in sepal-ui.

Every module content can be called directly from the package.

Example:
    .. jupyter-execute::

        from sepal_ui import aoi

        aoi.AoiTile(gee=False)
"""

from .aoi_model import *
from .aoi_tile import *
from .aoi_view import *



"""
Add extra behavior to radio and radio group to allow radio to be used as stand_alone components.

The radio from ipyvuetify cannot be used as stand alone elements like checkboxes (https://github.com/vuetifyjs/vuetify/issues/2345). Here the radio included in a radio group will have a v_model trait embedding the active status of the button. In short it's an overlay to reflect the JS into the Python code.
"""

import ipyvuetify as v
from traitlets import Bool, observe

from sepal_ui.sepalwidgets.sepalwidget import SepalWidget

__all__ = ["Radio", "RadioGroup"]


class Radio(v.Radio, SepalWidget):
    """Radio with extra active property."""

    active: Bool = Bool(allow_none=True).tag(sync=True)
    "True when active, False when not"


class RadioGroup(v.RadioGroup, SepalWidget):
    """
    This class will change the active of the included radio according to its v_model and vice-versa.
    """

    @observe("children")
    def link_radios(self, *args) -> None:
        """link all the radio children to the v_model."""
        for w in self.get_children(klass=Radio):
            w.observe(self.update_v_model, "active")

    def update_v_model(self, change: dict) -> None:
        """If a widget value is changed to True, update the v_model of the radioGroup."""
        if change["new"] is True:
            self.v_model = change["owner"].value

        return

    @observe("v_model")
    def update_radios(self, *args) -> None:
        """
        Update radios v_model.

        Change the v_model of every subsequent radios buttons according to new value of the radioGroup.
        children that are not Radios will be ignored.
        """
        for w in self.get_children(klass=Radio):
            w.active = w.value == self.v_model

        return



"""
Decorators used in sepal-ui.

used for multiple use-case sucha as (but not limited):
- catch errors in scripts to avoid Voila app freeze
- redirect error to a specific Alert object 
- Initialize EE
- debug widgets 
...
"""

import os
import warnings
from functools import wraps
from itertools import product
from pathlib import Path
from typing import Any, Callable, List, Union

import ee
import httplib2
import ipyvuetify as v
from deprecated.sphinx import versionadded

# from sepal_ui.scripts.utils import init_ee
from sepal_ui.scripts.warning import SepalWarning

################################################################################
# This method is a copy of the one from utils. It should stay there
# as long as there is deprecation warning in utils, we cannot import it due to a circular
# import. This method should then be removed in v3.0 when sd won't be imported by utils
#


def init_ee() -> None:
    """
    Initialize earth engine according to the environment.

    It will use the creddential file if the EARTHENGINE_TOKEN env variable exist.
    Otherwise it use the simple Initialize command (asking the user to register if necessary).
    """
    # only do the initialization if the credential are missing
    if not ee.data._credentials:

        # if the credentials token is asved in the environment use it
        if "EARTHENGINE_TOKEN" in os.environ:

            # write the token to the appropriate folder
            ee_token = os.environ["EARTHENGINE_TOKEN"]
            credential_folder_path = Path.home() / ".config" / "earthengine"
            credential_folder_path.mkdir(parents=True, exist_ok=True)
            credential_file_path = credential_folder_path / "credentials"
            credential_file_path.write_text(ee_token)

        # if the user is in local development the authentication should
        # already be available
        ee.Initialize(http_transport=httplib2.Http())

    return


################################################################################


@versionadded(version="3.0", reason="moved from utils to a dedicated module")
def catch_errors(alert: v.Alert, debug: bool = False) -> Any:
    """
    Decorator to execute try/except sentence and catch errors in the alert message.

    If debug is True then the error is raised anyway.

    Args:
        alert (sw.Alert): Alert to display errors
        debug (bool): Wether to raise the error or not, default to false

    Returns:
        The return statement of the decorated method
    """

    def decorator_alert_error(func):
        @wraps(func)
        def wrapper_alert_error(*args, **kwargs):
            value = None
            try:
                value = func(*args, **kwargs)
            except Exception as e:
                alert.add_msg(f"{e}", type_="error")
                if debug:
                    raise e
            return value

        return wrapper_alert_error

    return decorator_alert_error


@versionadded(version="3.0", reason="moved from utils to a dedicated module")
def need_ee(func: Callable) -> Any:
    """
    Decorator to execute check if the object require EE binding.

    Trigger an exception if the connection is not possible.

    Args:
        func: the object on which the decorator is applied

    Returns:
        The return statement of the decorated method
    """

    @wraps(func)
    def wrapper_ee(*args, **kwargs):

        # try to connect to ee
        try:
            init_ee()
        except Exception:
            raise Exception("This function needs an Earth Engine authentication")

        return func(*args, **kwargs)

    return wrapper_ee


@versionadded(version="3.0", reason="moved from utils to a dedicated module")
def loading_button(
    alert: Union[v.Alert, None] = None,
    button: Union[v.Btn, None] = None,
    debug: bool = False,
) -> Any:
    """
    Decorator to execute try/except sentence and toggle loading button object.

    Designed to work within the Tile object, or any object that have a self.btn and self.alert set.

    Args:
        button: Toggled button
        alert: the alert to display the error message
        debug: wether or not the exception should stop the execution. default to False

    Returns:
        The return statement of the decorated method
    """

    def decorator_loading(func):
        @wraps(func)
        def wrapper_loading(self, *args, **kwargs):

            # set btn and alert
            # Change name of variable to assign it again in this scope
            button_ = self.btn if not button else button
            alert_ = self.alert if not alert else alert

            # Clean previous loaded messages in alert
            alert_.reset()

            button_.toggle_loading()  # Start loading
            value = None
            try:
                # Catch warnings in the process function
                with warnings.catch_warnings(record=True) as w_list:
                    value = func(self, *args, **kwargs)

                # Check if there are warnings in the function and append them
                # Use append msg as several warnings could be triggered
                if w_list:

                    # split the warning list
                    w_list_sepal = [
                        w for w in w_list if isinstance(w.message, SepalWarning)
                    ]

                    # display the sepal one
                    ms_list = [
                        f"{w.category.__name__}: {w.message.args[0]}"
                        for w in w_list_sepal
                    ]
                    [alert_.append_msg(ms, type_="warning") for ms in ms_list]

                    # only display them in the console if debug mode
                    if debug:

                        def custom_showwarning(w):
                            return warnings.showwarning(
                                message=w.message,
                                category=w.category,
                                filename=w.filename,
                                lineno=w.lineno,
                                line=w.line,
                            )

                        [custom_showwarning(w) for w in w_list]

            except Exception as e:
                alert_.add_msg(f"{e}", "error")
                if debug:
                    button_.toggle_loading()  # Stop loading button if there is an error
                    raise e

            button_.toggle_loading()  # Stop loading button

            return value

        return wrapper_loading

    return decorator_loading


@versionadded(version="3.0", reason="moved from utils to a dedicated module")
def switch(
    *params, debug: bool = True, on_widgets: List[str] = [], targets: List[bool] = []
) -> Any:
    """
    Decorator to switch the state of input boolean parameters on class widgets or the class itself.

    If on_widgets is defined, it will switch the state of every widget
    parameter, otherwise it will change the state of the class (self). You can also set
    two decorators on the same function, one could affect the class and other the widgets.

    Args:
        *params: any boolean parameter of a SepalWidget.
        debug: Whether trigger or not an Exception if the decorated function fails.
        on_widgets: List of widget names into the class
        targets: list of the target value (value taht will be set on switch. default to the inverse of the current state.

    Returns:
        The return statement of the decorated method
    """

    def decorator_switch(func):
        @wraps(func)
        def wrapper_switch(self, *args, **kwargs):

            widgets_len = len(on_widgets)
            targets_len = len(targets)

            # sanity check on targets and on_widgets
            if widgets_len and targets_len:
                if widgets_len != targets_len:
                    raise IndexError(
                        f'the length of "on_widgets" ({widgets_len}) is different from the length of "targets" ({targets_len})'
                    )

            # create the list of target values based on the target list
            # or the initial values of the widgets params
            # The first one is taken as reference
            if not targets_len:
                w = getattr(self, on_widgets[0]) if widgets_len else self
                targets_ = [bool(getattr(w, p)) for p in params]
            else:
                targets_ = targets

            if widgets_len:

                # Verify that the input elements are strings
                wrong_types = [
                    (w, type(w)) for w in on_widgets if not isinstance(w, str)
                ]

                if len(wrong_types):
                    errors = [
                        f"Received:{w_type} for widget: {w}."
                        for w, w_type in wrong_types
                    ]

                    raise TypeError(
                        f"All on_widgets list elements has to be strings. [{' '.join(errors)}]"
                    )

                missing_widgets = [w for w in on_widgets if not hasattr(self, w)]

                if missing_widgets:
                    raise Exception(
                        f"The provided {missing_widgets} widget(s) does not exist in the current class"
                    )

                def w_assign(bool_targets):

                    params_targets = [
                        (p, bool_targets[i]) for i, p in enumerate(params)
                    ]

                    for (w_name, p_t) in product(on_widgets, params_targets):
                        param, target = p_t
                        widget = getattr(self, w_name)
                        setattr(widget, param, target)

            else:

                def w_assign(bool_targets):
                    for i, p in enumerate(params):
                        setattr(self, p, bool_targets[i])

            # assgn the parameters to the target inverse
            w_assign([not t for t in targets_])

            # execute the function and catch errors
            try:
                func(self, *args, **kwargs)

            except Exception as e:
                if debug:
                    w_assign(targets_)
                    raise e

            # reassign the parameters to the targets
            w_assign(targets_)

        return wrapper_switch

    return decorator_switch



"""
Customized ``Control`` to zoom in and out on the map.
"""
from typing import Optional

from ipyleaflet import Map, WidgetControl

from sepal_ui import sepalwidgets as sw
from sepal_ui.mapping.map_btn import MapBtn


class ZoomControl(WidgetControl):

    plus: Optional[MapBtn] = None
    "The plus btn"

    minus: Optional[MapBtn] = None
    "the minus btn"

    m: Optional[Map] = None
    "the map to manipulate"

    def __init__(self, m: Map, **kwargs) -> None:
        """
        Customized ``Control`` to zoom in and out on the map.

        Replace the built-in zoom control of ipyleaflet to match the theme of sepal-ui based applications. It is by default positioned in the top-right corner

        Args:
            m: the map to manipulate
            kwargs: any ``ipyleaflet.widgetControl`` arguments
        """
        # init the map
        self.m = m

        # create the 2 btns
        self.plus = MapBtn("fa-solid fa-plus", _metadata={"step": 1})
        self.minus = MapBtn("fa-solid fa-minus", _metadata={"step": -1})

        # customize their layout by removing the bottom radius
        self.plus.class_list.add("v-zoom-plus")
        self.minus.class_list.add("v-zoom-minus")

        # agregate them in a layout
        content = sw.Layout(column=True, children=[self.plus, self.minus])

        # by default Zoomcontrols will be displayed in the topright corner
        kwargs.setdefault("position", "topright")
        kwargs["widget"] = content

        super().__init__(**kwargs)

        # add js behaviour
        self.plus.on_event("click", self.zoom)
        self.minus.on_event("click", self.zoom)

    def zoom(self, widget: MapBtn, *args) -> None:
        """update the zoom according to the clicked btn."""
        # read min and max zoom
        max_zoom = self.m.max_zoom or 24
        min_zoom = self.m.min_zoom or 0

        # computed zoom
        zoom = self.m.zoom + widget._metadata["step"]

        # adapt to the limitations of the map
        self.m.zoom = max(min(zoom, max_zoom), min_zoom)

        return



#!/usr/bin/python3

"""
Script to create the skeleton of a sepal-ui module.

The script will extract the skeleton of a module from the sepal_ui_template GitHub repository. This template will be adapted based on the answsers to the questionnaire.
Placeholdre from the template will be replaced and the directory will be synced with a GitHub freshly created repository. Note that the repository need to be fully empty when the command is launched.
"""

import argparse
import json
import re
import subprocess
from distutils.util import strtobool
from pathlib import Path

from colorama import Fore, init

# init colors for all plateforms
init()

# init parser
parser = argparse.ArgumentParser(description=__doc__, usage="module_factory")


def set_default_readme(
    folder: Path, module_name: str, description: str, url: str
) -> None:
    """
    Write a default README.md file and overwrite the existing one.

    Args:
        folder: the module directory
        module_name: the module name used as title everywhere
        description: the description of the module
        url: the url of the module repository in GitHub
    """
    print("Write a default README.md file")

    license = f"{url}/blob/master/LICENSE"

    file = folder / "README.md"
    # write_text cannot handle append
    with file.open("w") as readme:
        readme.writelines(
            [
                f"# {module_name}\n",
                "\n",
                f"[![License: MIT](https://img.shields.io/badge/License-MIT-yellow.svg)]({license})\n",
                "[![Black badge](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n",
                "\n",
                "# About\n",
                "\n",
                f"{description}\n",
            ]
        )

    return


def set_default_about(folder: Path, description: str) -> None:
    """
    Write a default ABOUT_en.md file and overwrite the existing one.

    Args:
        folder: the directory of the module
        description: the description of the module functions
    """
    print("Write a default ABOUT_en.md file")

    # creating the dir
    dir_ = folder / "utils"
    dir_.mkdir(exist_ok=True)

    # set the file
    file = dir_ / "ABOUT_en.md"

    with file.open("w") as about:
        about.write(f"{description}  \n")

    return


def set_module_name(folder: Path, module_name: str) -> None:
    """
    Use the module name in the different translation dictionaries.

    Args:
        folder: the directory of the module
        module_name: the module name
    """
    print("Update the module name in the json translation dictionaries")

    # edit the en file
    app_file = folder / "component" / "message" / "en" / "app.json"

    with app_file.open("r") as f:
        data = json.load(f)

    data["app"]["title"] = module_name

    with app_file.open("w") as f:
        json.dump(data, f, indent=4)

    return


def set_module_name_doc(folder: Path, url: str, module_name: str) -> None:
    """
    Set the module name in each documentation file and set the appropriate repository in the link.

    Args:
        folder: the directory of the module
        url: the url of the GitHub repository
        module_name: the module name
    """
    # get the documentation folder
    doc_dir = folder / "doc"

    # loop in the available languages
    for file in doc_dir.glob("*.rst"):

        with file.open() as f:
            text = f.read()

        text = text.replace("Module_name", module_name)
        text = text.replace("===========", "=" * len(module_name))
        text = text.replace("https://github.com/12rambau/sepal_ui", url)

        with file.open("w") as f:
            f.write(text)
    return


def set_drawer_link(folder: Path, url: str) -> None:
    """
    Replace the reference to the default repository to the one provided by the user.

    Args:
        folder: the directory of the module
        url: the url of the GitHub repository
    """
    print("Update the drawers link with the new repository one")

    # get the ui file
    ui = folder / "ui.ipynb"

    # read the file
    with ui.open() as f:
        ui_content = f.read()

    # replace the target strings
    ui_content = ui_content.replace("https://github.com/12rambau/sepal_ui", url)

    # write everything down again
    with ui.open("w") as f:
        f.write(ui_content)

    return


def main() -> None:
    """Launch the process."""
    # parse agruments
    parser.parse_args()

    # welcome the user
    print(f"{Fore.YELLOW}sepal-ui module factory{Fore.RESET}")

    print("Initializing module creation by setting up your module parameters")
    print("‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾")

    # ask the name of the module
    module_name = input(f"{Fore.CYAN}Provide a module name: \n{Fore.RESET}")
    if not module_name:
        raise Exception(f"{Fore.RED}A module name should be set")
    question = f"{Fore.CYAN}Provide the URL of an empty github repository: \n{Fore.RESET}"  # fmt: skip
    github_url = input(question)
    if not github_url:
        msg = f"{Fore.RED}A module name should be set with an asociated github repository"  # fmt: skip
        raise Exception(msg)
    question = f"{Fore.CYAN}Provide a short description for your module (optional): \n{Fore.RESET}"  # fmt: skip
    description = input(question)

    # default to a panel application
    question = f"{Fore.CYAN}Do you need a fullscreen application [n]? \n{Fore.RESET}"
    type_ = input(question) or "no"
    branch = "map_app" if strtobool(type_) is True else "panel_app"

    # adapt the name of the module to remove any special characters and spaces
    normalized_name = re.sub("[^a-zA-Z\d\-\_]", "_", module_name)

    print("Build the module init configuration")
    print("‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾")

    # clone the repository in a folder that has the normalized module name
    folder = Path.cwd() / normalized_name

    template_dir = Path(__file__).parents[1] / "templates" / branch / "."
    subprocess.run(["cp", "-r", str(template_dir), str(folder)], cwd=Path.cwd())

    # replace the placeholders
    url = github_url.replace(".git", "").replace(
        "git@github.com:", "https://github.com/"
    )

    set_default_readme(folder, module_name, description, url)
    set_default_about(folder, description)
    set_module_name(folder, module_name)
    set_drawer_link(folder, url)
    set_module_name_doc(folder, url, module_name)

    default = "main"

    # execute the creation process in command lines
    # - init the new git repository
    # - change default branch name if using an old version of git
    # - add the configuration of the git repository
    # - add all the files in the git repo
    # - run all the precommit at least once
    # - include the correction from pre-commits
    # - first commit
    # - create a branch
    # - add the remote
    # - make the first push
    # - create a release branch and push it to the server
    # - checkout to master
    subprocess.run(["git", "init"], cwd=folder)
    subprocess.run(["git", "checkout", "-b", default], cwd=folder)
    subprocess.run(["pre-commit", "install"], cwd=folder)
    subprocess.run(["git", "add", "."], cwd=folder)
    subprocess.run(["pre-commit", "run", "--all-files"], cwd=folder)
    subprocess.run(["git", "add", "."], cwd=folder)
    subprocess.run(["git", "commit", "-m", "initial commit"], cwd=folder)
    subprocess.run(["git", "branch", "-M", default], cwd=folder)
    subprocess.run(["git", "remote", "add", "origin", str(github_url)], cwd=folder)
    subprocess.run(["git", "push", "-u", "origin", default], cwd=folder)
    subprocess.run(["git", "checkout", "-b", "release"], cwd=folder)
    subprocess.run(["git", "push", "--set-upstream", "origin", "release"], cwd=folder)
    subprocess.run(["git", "checkout", default], cwd=folder)

    # exit message
    print(
        f"{Fore.YELLOW}Have a look to the git command executed in the process. if any of them is displaying an error, the final folder may not have been created{Fore.RESET}"
    )
    print(
        f"{Fore.YELLOW}If that's the case, delete the folder in your sepal instance (if there is any) and start the process again or contact us via github issues.{Fore.RESET}"
    )
    print(f"{Fore.GREEN}You created a new module named: {module_name}{Fore.RESET}")
    print(
        f"{Fore.GREEN}You can find its code in {folder} inside your sepal environment.{Fore.RESET}"
    )
    print()
    print("Let's code !")


if __name__ == "__main__":
    main()



"""
sepal-ui configuration to save in the user machine.

To make applications more user-friendly sepal-ui save few parameters in a config file. 
Currently we save the language and the theme so that it won't need to be reset every time. 
"""

from configparser import ConfigParser
from pathlib import Path

config_file = Path.home() / ".sepal-ui-config"
config = ConfigParser()
"ConfigParser: the configuration object generated by sepal-ui based application to save parameters such as language or theme"
config.read(config_file)

# Set a default conifg when there is not file
if not config_file.exists():
    config.add_section("sepal-ui")
    config.set("sepal-ui", "theme", "dark")
    config.set("sepal-ui", "theme", "en")
    config.write(config_file.open("w"))



"""
``Card`` object dedicated to AOI selection. It does not include maps.
"""

from datetime import datetime as dt
from pathlib import Path
from typing import Dict, List, Optional, Union

import ipyvuetify as v
import pandas as pd
import traitlets as t
from deprecated.sphinx import versionadded
from typing_extensions import Self

import sepal_ui.sepalwidgets as sw
from sepal_ui import mapping as sm
from sepal_ui.aoi.aoi_model import AoiModel
from sepal_ui.message import ms
from sepal_ui.scripts import decorator as sd
from sepal_ui.scripts import utils as su

CUSTOM = AoiModel.CUSTOM
ADMIN = AoiModel.ADMIN
ALL = "All"
select_methods = AoiModel.METHODS

__all__ = ["AoiView", "select_methods"]


class MethodSelect(sw.Select):
    def __init__(
        self,
        methods: Union[str, List[str]] = "ALL",
        gee: bool = True,
        map_: Optional[sm.SepalMap] = None,
    ) -> None:
        """
        A method selector.

        It will list the available methods for this very AoiView.
        'ALL' will select all the available methods (default)
        'ADMIN' only the admin one, 'CUSTOM' only the custom one.
        'XXX' will add the selected method to the list when '-XXX' will discard it.
        You cannot mix adding and removing behaviours.

        Args:
            methods: a list of methods from the available list (ADMIN0, ADMIN1, ADMIN2, SHAPE, DRAW, POINTS, ASSET)
            map_: link the aoi_view to a custom SepalMap to display the output, default to None
            gee: wether to bind to ee or not
        """
        # create the method list
        if methods == "ALL":
            self.methods = select_methods
        elif methods == "ADMIN":
            self.methods = {
                k: v for k, v in select_methods.items() if v["type"] == ADMIN
            }
        elif methods == "CUSTOM":
            self.methods = {
                k: v for k, v in select_methods.items() if v["type"] == CUSTOM
            }
        elif type(methods) == list:

            if any(m[0] == "-" for m in methods) != all(m[0] == "-" for m in methods):
                raise Exception("You mixed adding and removing, punk")

            if methods[0][0] == "-":

                to_remove = [method[1:] for method in methods]

                # Rewrite the methods instead of mutate the class methods
                self.methods = {
                    k: v for k, v in select_methods.items() if k not in to_remove
                }

            else:
                self.methods = {k: select_methods[k] for k in methods}
        else:
            raise Exception("I don't get what you meant")

        # clean the list from things we can't use
        gee is True or self.methods.pop("ASSET", None)
        map_ is not None or self.methods.pop("DRAW", None)

        # build the item list with header
        prev_type = None
        items = []
        for k, m in self.methods.items():
            current_type = m["type"]

            if prev_type != current_type:
                items.append({"header": current_type})
            prev_type = current_type

            items.append({"text": m["name"], "value": k})

        # create the input
        super().__init__(label=ms.aoi_sel.method, items=items, v_model="", dense=True)


class AdminField(sw.Select):

    gee: bool = True
    "wether or not to depend on earthengine"

    level: int = -1
    "The admin level of the current field"

    parent: Optional[sw.Select] = None
    "The parent adminfield object"

    def __init__(
        self, level: int, parent: Optional[sw.Select] = None, gee: bool = True
    ) -> None:
        """
        An admin level selector.

        It is binded to ee (GAUL 2015) or not (GADM 2021). allows to select administrative codes taking into account the administrative parent code and displaying humanly readable administrative names.

        Args:
            level: The administrative level of the field
            parent: the adminField that deal with the parent admin level of the current selector. used to narrow down the possible options
            ee: wether to use ee or not (default to True)
        """
        # save ee state
        self.gee = gee

        # get the level info
        self.level = level
        self.parent = parent

        # init an empty widget
        super().__init__(
            v_model=None, items=[], clearable=True, label=ms.aoi_sel.adm[level]
        )

        # add js behaviour
        self.parent is None or self.parent.observe(self._update, "v_model")

    def show(self) -> Self:
        """
        when an admin field is shown, show its parent as well.
        """
        super().show()
        self.parent is None or self.parent.show()

        return self

    def get_items(self, filter_: str = "") -> Self:
        r"""
        Update the item list based on the given filter.

        Args:
            filter\_ (str): The code of the parent v_model to filter the current results
        """
        # extract the level list
        df = (
            pd.read_parquet(AoiModel.FILE[self.gee])
            .astype(str)
            .drop_duplicates(subset=AoiModel.CODE[self.gee].format(self.level))
            .sort_values(AoiModel.NAME[self.gee].format(self.level))
        )

        # filter it
        if filter_:
            df = df[df[AoiModel.CODE[self.gee].format(self.level - 1)] == filter_]

        # formatted as a item list for a select component
        self.items = [
            {
                "text": su.normalize_str(
                    r[AoiModel.NAME[self.gee].format(self.level)], folder=False
                ),
                "value": r[AoiModel.CODE[self.gee].format(self.level)],
            }
            for _, r in df.iterrows()
        ]

        return self

    def _update(self, change: dict) -> Self:
        """update the item list of the admin select."""
        # reset v_model
        self.v_model = None

        # update the items list
        if change["new"]:
            self.get_items(change["new"])

        return self


class AoiView(sw.Card):

    # ##########################################################################
    # ###                             widget parameters                      ###
    # ##########################################################################

    updated: t.Int = t.Int(0).tag(sync=True)
    "Traitlets triggered every time a AOI is selected"

    gee: bool = True
    "Either or not he aoi_view is connected to gee"

    folder: Union[str, Path] = ""
    "The folder name used in GEE related component, mainly used for debugging"

    model: Optional[AoiModel] = None
    "The model to create the AOI from the selected parameters"

    map_style: Optional[dict] = None
    "The predifined style of the aoi on the map"

    # ##########################################################################
    # ###                            the embeded widgets                     ###
    # ##########################################################################

    map_: Optional[sm.SepalMap] = None
    "The map to draw the AOI"

    aoi_dc: Optional[sm.DrawControl] = None
    "the drawing control associated with DRAW method"

    w_method: Optional[MethodSelect] = None
    "The widget to select the method"

    components: Dict[str, v.VuetifyWidget] = {}
    "The followingwidgets used to define AOI"

    w_admin_0: Optional[AdminField] = None
    "The widget used to select admin level 0"

    w_admin_1: Optional[AdminField] = None
    "The widget used to select admin level 1"

    w_admin_2: Optional[AdminField] = None
    "The widget used to select admin level 2"

    w_vector: Optional[sw.VectorField] = None
    "The widget used to select vector shapes"

    w_points: Optional[sw.LoadTableField] = None
    "The widget used to select points files"

    w_draw: Optional[sw.TextField] = None
    "The widget used to select the name of a drawn shape (only if :code:`map_ != None`)"

    w_asset: Optional[sw.AssetSelect] = None
    "The widget used to select asset name of a featureCollection (only if :code:`gee == True`)"

    btn: Optional[sw.Btn] = None
    "A default btn"

    alert: Optional[sw.Alert] = None
    "A alert to display message to the end user"

    @versionadded(
        version="2.11.3",
        reason="Model is now an optional parameter to AoiView, it can be created from outside and passed to the initialization function.",
    )
    def __init__(
        self,
        methods: Union[str, List[str]] = "ALL",
        map_: Optional[sm.SepalMap] = None,
        gee: bool = True,
        folder: Union[str, Path] = "",
        model: Optional[AoiModel] = None,
        map_style: Optional[dict] = None,
        **kwargs,
    ) -> None:
        r"""
        Versatile card object to deal with the aoi selection. multiple selection method are available (see the MethodSelector object) and the widget can be fully customizable. Can also be bound to ee (ee==True) or not (ee==False).

        Args:
            methods: the methods to use in the widget, default to 'ALL'. Available: {'ADMIN0', 'ADMIN1', 'ADMIN2', 'SHAPE', 'DRAW', 'POINTS', 'ASSET', 'ALL'}
            map\_: link the aoi_view to a custom SepalMap to display the output, default to None
            gee: wether to bind to ee or not
            vector: the path to the default vector object
            admin: the administrative code of the default selection. Need to be GADM if :code:`ee==False` and GAUL 2015 if :code:`ee==True`.
            asset: the default asset. Can only work if :code:`ee==True`
            map_style: the predifined style of the aoi. It's by default using a "success" ``sepal_ui.color`` with 0.5 transparent fill color. It can be completly replace by a fully qualified `style dictionnary <https://ipyleaflet.readthedocs.io/en/latest/layers/geo_json.html>`__. Use the ``sepal_ui.color`` object to define any color to remain compatible with light and dark theme.
        """
        # set ee dependencie
        self.gee = gee
        self.folder = folder
        if gee is True:
            su.init_ee()

        # get the model
        self.model = model or AoiModel(gee=gee, folder=folder, **kwargs)

        # get the map if filled
        self.map_ = map_

        # get the aoi geoJSON style
        self.map_style = map_style

        # create the method widget
        self.w_method = MethodSelect(methods, gee=gee, map_=map_)

        # add the methods blocks
        self.w_admin_0 = AdminField(0, gee=gee).get_items()
        self.w_admin_1 = AdminField(1, self.w_admin_0, gee=gee)
        self.w_admin_2 = AdminField(2, self.w_admin_1, gee=gee)
        self.w_vector = sw.VectorField(label=ms.aoi_sel.vector)
        self.w_points = sw.LoadTableField(label=ms.aoi_sel.points)

        # group them together with the same key as the select_method object
        self.components = {
            "ADMIN0": self.w_admin_0,
            "ADMIN1": self.w_admin_1,
            "ADMIN2": self.w_admin_2,
            "SHAPE": self.w_vector,
            "POINTS": self.w_points,
        }

        # hide them all
        [c.hide() for c in self.components.values()]

        # use the same alert as in the model
        self.alert = sw.Alert()

        # bind the widgets to the model
        (
            self.model.bind(self.w_admin_0, "admin")
            .bind(self.w_admin_1, "admin")
            .bind(self.w_admin_2, "admin")
            .bind(self.w_vector, "vector_json")
            .bind(self.w_points, "point_json")
            .bind(self.w_method, "method")
        )

        # defint the asset select separately. If no gee is set up we don't want any
        # gee based widget to be requested. If it's the case, application that does not support GEE
        # will crash if the user didn't authenticate
        if self.gee:
            self.w_asset = sw.VectorField(
                label=ms.aoi_sel.asset, gee=True, folder=self.folder, types=["TABLE"]
            )
            self.w_asset.hide()
            self.components["ASSET"] = self.w_asset
            self.model.bind(self.w_asset, "asset_json")

        # define DRAW option separately as it will only work if the map is set
        if self.map_:
            self.w_draw = sw.TextField(label=ms.aoi_sel.aoi_name).hide()
            self.components["DRAW"] = self.w_draw
            self.model.bind(self.w_draw, "name")
            self.aoi_dc = sm.DrawControl(self.map_)
            self.aoi_dc.hide()

        # add a validation btn
        self.btn = sw.Btn(ms.aoi_sel.btn)

        # create the widget
        self.children = (
            [self.w_method] + [*self.components.values()] + [self.btn, self.alert]
        )

        super().__init__(**kwargs)

        # js events
        self.w_method.observe(self._activate, "v_model")  # activate widgets
        self.btn.on_event("click", self._update_aoi)  # load the informations

        # reset te aoi_model
        self.model.clear_attributes()

    @sd.loading_button(debug=True)
    def _update_aoi(self, *args) -> Self:
        """load the object in the model & update the map (if possible)."""
        # read the information from the geojson datas
        if self.map_:
            self.model.geo_json = self.aoi_dc.to_json()

        # update the model
        self.model.set_object()
        self.alert.add_msg(ms.aoi_sel.complete, "success")

        # update the map
        if self.map_:
            self.map_.remove_layer("aoi", none_ok=True)
            self.map_.zoom_bounds(self.model.total_bounds())
            self.map_.add_layer(self.model.get_ipygeojson(self.map_style))

            self.aoi_dc.hide()

        # tell the rest of the apps that the aoi have been updated
        self.updated += 1

        return self

    def reset(self) -> Self:
        """clear the aoi_model from input and remove the layer from the map (if existing)."""
        # reset the view of the widgets
        self.w_method.v_model = None

        # clear the map
        if self.map_ is not None:
            self.map_.remove_layer("aoi", none_ok=True)

        return self

    @sd.switch("loading", on_widgets=["w_method"])
    def _activate(self, change: dict) -> None:
        """activate the adapted widgets."""
        # clear and hide the alert
        self.alert.reset()

        # hide the widget so that the user doens't see status changes
        [w.hide() for w in self.components.values()]

        # clear the inputs in a second step as reseting a FileInput can be long
        [w.reset() for w in self.components.values()]

        # deactivate or activate the dc
        # clear the geo_json saved features to start from scratch
        if self.map_:
            if change["new"] == "DRAW":
                self.aoi_dc.show()
            else:
                self.aoi_dc.hide()

        # activate the correct widget
        w = next((w for k, w in self.components.items() if k == change["new"]), None)
        w is None or w.show()

        # init the name to the current value
        now = dt.now().strftime("%Y-%m-%d_%H-%M-%S")
        self.w_draw.v_model = None if change["new"] is None else f"Manual_aoi_{now}"

        return



#!/usr/bin/python3

"""
Script to add a testing kernel in the local environment.

This script should be run in a module folder. From there, it will create a  new venv in the "module_venv" folder.
This venv will be named after the module repository name. All librairies specified in the requirements.txt file will be installed.
Then the script will read the "module.yaml" file (if existing) and exctract, the name of the module and the entry_point file (ususally "ui.ipynb").
The venv will be added to the list of usable Jupyter kernel. It will be named: "test-<name_of_the_repo>" and displayed as "(test) <name of the module>".
The kernel should automatically be added to the entry_point file.
"""

import argparse
import json
import subprocess
from pathlib import Path
from shutil import rmtree

import yaml
from colorama import Fore, init

# init colors for all plateforms
init()

# init parser
parser = argparse.ArgumentParser(description=__doc__, usage="module_venv")


def main() -> None:
    """Launch the venv creation process."""
    # read arguments (there should be none)
    parser.parse_args()

    # welcome the user
    print(f"{Fore.YELLOW}venv creation interface{Fore.RESET}")

    # check that the local folder is a module folder
    ui_file = Path.cwd() / "ui.ipynb"
    if not ui_file.is_file():
        raise Exception(f"{Fore.RED}This is not a module folder.")

    # create a venv folder
    print('create the venv directory: "module-venv"')
    venv_dir = Path.home() / "module-venv"
    venv_dir.mkdir(exist_ok=True)

    # create a venv folder associated with the current repository
    print(f'create a venv directory for the current app: "{Path.cwd().name}"')
    current_dir_venv = venv_dir / Path.cwd().name

    # empty the folder from anything already in there
    # equivalement to flushing the existing venv (it's just faster than pip)
    if current_dir_venv.is_dir():
        rmtree(current_dir_venv)

    current_dir_venv.mkdir(exist_ok=True)

    # init the venv
    print("Initializing the venv, this process will take some minutes.")

    subprocess.run(["python3", "-m", "venv", str(current_dir_venv)], cwd=Path.cwd())

    # extract python and pip from thevenv
    pip = current_dir_venv / "bin" / "pip"
    python3 = current_dir_venv / "bin" / "python3"

    for lib in ["wheel", "Cython", "ipykernel"]:
        subprocess.run([str(pip), "install", lib], cwd=Path.cwd())

    # install all the requirements
    req = Path.cwd() / "requirements.txt"
    subprocess.run([str(pip), "install", "-r", str(req)], cwd=Path.cwd())

    # search for the module.yaml file
    # it embeds name and entry point
    module_config = Path().cwd() / "module.yaml"

    if module_config.is_file():
        with module_config.open() as f:
            data = yaml.safe_load(f)
        entry_point = Path.cwd() / data["entry_point"]
        name = data["name"]

    else:
        entry_point = Path.cwd() / "ui.ipynb"
        name = Path.cwd().stem

    # create the kernel from venv
    name = f"test-{Path.cwd().stem}"
    display_name = f"(test) {name}"
    subprocess.run(
        [
            str(python3),
            "-m",
            "ipykernel",
            "install",
            "--user",
            "--name",
            name,
            "--display-name",
            display_name,
        ],
        cwd=Path.cwd(),
    )

    # change the kernel of the entrypoint to use this one instead
    with entry_point.open() as f:
        data = json.load(f)

    data["metadata"]["kernelspec"]["display_name"] = display_name
    data["metadata"]["kernelspec"]["name"] = name

    entry_point.write_text(json.dumps(data, indent=1))

    # display last message to the end user
    print(
        f'{Fore.GREEN}The test venv have been created, it can be found in the kernel list as "{display_name}". It has automatically been added to the entry point of the application: {entry_point.name}.{Fore.RESET}'
    )

    return


if __name__ == "__main__":
    main()



"""
All the helper function of sepal-ui. 
"""

import math
import os
import random
import re
import string
import warnings
from pathlib import Path
from typing import Any, Sequence, Union
from urllib.parse import urlparse

import ee
import httplib2
import ipyvuetify as v
from deprecated.sphinx import deprecated, versionadded
from matplotlib import colors as c
from unidecode import unidecode

import sepal_ui
from sepal_ui.conf import config, config_file
from sepal_ui.message import ms
from sepal_ui.scripts import decorator as sd
from sepal_ui.scripts.warning import SepalWarning


def hide_component(widget: v.VuetifyWidget) -> v.VuetifyWidget:
    """
    hide a vuetify based component.

    Args:
        widget: the widget to hide

    Returns:
        the widget
    """
    if isinstance(widget, sepal_ui.sepalwidgets.sepalwidget.SepalWidget):
        widget.hide()

    elif "d-none" not in str(widget.class_):
        widget.class_ = str(widget.class_).strip() + " d-none"

    return widget


def show_component(widget: v.VuetifyWidget) -> v.VuetifyWidget:
    """
    show a vuetify based component.

    Args:
        widget: the widget to hide

    Returns:
        the widget
    """
    if isinstance(widget, sepal_ui.sepalwidgets.sepalwidget.SepalWidget):
        widget.show()

    elif "d-none" in str(widget.class_):
        widget.class_ = widget.class_.replace("d-none", "")

    return widget


def create_download_link(pathname: Union[str, Path]) -> str:
    """
    Create a clickable link to download the pathname target.

    Args:
        pathname: the pathname th download

    Returns:
        the download link
    """
    # return the link if it's an absolute url
    if isinstance(pathname, str) and bool(urlparse(str(pathname)).netloc):
        return pathname

    # create a downloadable link from the jupyter node
    pathname = Path(pathname)
    try:
        download_path = pathname.relative_to(Path.home())
    except ValueError:
        download_path = pathname

    # I want to use the ipyurl lib to guess the url of the Jupyter server on the fly
    # but I don't really understand how it works
    # so here is an ugly fix only compatible with SEPAL
    link = f"https://sepal.io/api/sandbox/jupyter/files/{download_path}"

    return link


def random_string(string_length: int = 3) -> str:
    """
    Generates a random string of fixed length.

    Args:
        string_length: Fixed length. Defaults to 3.

    Returns:
        A random string
    """
    letters = string.ascii_lowercase

    return "".join(random.choice(letters) for i in range(string_length))


def get_file_size(filename: Union[str, Path]) -> str:
    """
    Get the file size as string of 2 digit in the adapted scale (B, KB, MB....).

    Args:
        filename: the path to the file to mesure

    Returns:
        the file size in a readable humanly readable
    """
    file_size = Path(filename).stat().st_size

    if file_size == 0:
        return "0B"

    size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")

    i = int(math.floor(math.log(file_size, 1024)))
    s = file_size / (1024**i)

    return "{:.1f} {}".format(s, size_name[i])


def init_ee() -> None:
    """
    Initialize earth engine according to the environment.

    It will use the creddential file if the EARTHENGINE_TOKEN env variable exist.
    Otherwise it use the simple Initialize command (asking the user to register if necessary).
    """
    # only do the initialization if the credential are missing
    if not ee.data._credentials:

        # if the credentials token is asved in the environment use it
        if "EARTHENGINE_TOKEN" in os.environ:

            # write the token to the appropriate folder
            ee_token = os.environ["EARTHENGINE_TOKEN"]
            credential_folder_path = Path.home() / ".config" / "earthengine"
            credential_folder_path.mkdir(parents=True, exist_ok=True)
            credential_file_path = credential_folder_path / "credentials"
            credential_file_path.write_text(ee_token)

        # if the user is in local development the authentication should
        # already be available
        ee.Initialize(http_transport=httplib2.Http())

    return


def normalize_str(msg: str, folder: bool = True) -> str:
    """
    Normalize an str to make it compatible with file naming (no spaces, special chars ...etc).

    Params:
        msg: the string to sanitise
        folder: if the name will be used for folder naming or for display. if display, <'> and < > characters will be kept

    Returns:
        the modified str
    """
    regex = "[^a-zA-Z\d\-\_]" if folder else "[^a-zA-Z\d\-\_\ ']"

    return re.sub(regex, "_", unidecode(msg))


def to_colors(
    in_color: Union[str, Sequence], out_type: str = "hex"
) -> Union[str, tuple]:
    """
    Transform any color type into a color in the specified output format.

    Avalable format: [hex]

    Args:
        in_color: It can be a string (e.g., 'red', '#ffff00', 'ffff00') or RGB tuple (e.g., (255, 127, 0)).
        out_type: the type of the output color from ['hex']. default to 'hex'

    Returns:
        The color in the specified format. default to black.
    """
    # list of the color function used for the translation
    c_func = {"hex": c.to_hex}
    transform = c_func[out_type]

    out_color = "#000000"  # default black color

    if isinstance(in_color, tuple) and len(in_color) == 3:

        # rescale color if necessary
        if all(isinstance(item, int) for item in in_color):
            in_color = [c / 255.0 for c in in_color]

        return transform(in_color)

    else:

        # try to guess the color system
        try:
            return transform(in_color)
        except Exception:
            pass

        # try again by adding an extra # (GEE handle hex codes without #)
        try:
            return transform(f"#{in_color}")
        except Exception:
            pass

    return transform(out_color)


def next_string(string: str) -> str:
    """
    Create a string followed by an underscore and a consecutive number.

    Args:
        string: the initial string

    Returns:
        the incremented string
    """
    # if the string is already numbered the last digit is separeted from the rest of the string by an "_"
    split = string.split("_")
    end = split[-1]

    if end.isdigit():
        string = "_".join(split[:-1]) + f"_{int(end)+1}"
    else:
        string += "_1"

    return string


def set_config(key: str, value: str, section: str = "sepal-ui") -> None:
    """
    Set a config variable.

    Set the provided value to the given key for the given section in the sepal-ui config
    file.

    Args:
        key: key configuration name
        value: value to be referenced by the configuration key
        section: configuration section, defaults to sepal-ui.
    """
    # set the section if needed
    if "sepal-ui" not in config.sections():
        config.add_section(section)

    # set the value
    config.set("sepal-ui", key, value)

    # save back the file
    config.write(config_file.open("w"))

    return


@deprecated(
    version="2.9.1", reason="This function will be removed in favor of set_config()"
)
def set_config_locale(locale: str) -> None:
    """
    Set the provided local in the sepal-ui config file.

    Args:
        locale: a locale name in IETF BCP 47 (no verifications are performed)
    """
    return set_config("locale", locale)


@deprecated(
    version="2.9.1", reason="This function will be removed in favor of set_config()"
)
def set_config_theme(theme: str) -> None:
    """
    Set the provided theme in the sepal-ui config file.

    Args:
        theme: a theme name (currently supporting "dark" and "light")
    """
    return set_config("theme", theme)


@versionadded(version="2.7.1")
def set_type(color: str) -> str:
    r"""
    Return a pre-defined material colors based on the requested type\_ parameter.

    If the parameter is not a predifined color, fallback to "info" and will raise a warning. the colors can only be selected from ["primary", "secondary", "accent", "error", "info", "success", "warning", "anchor"].

    Args:
        color: the requested color

    Returns:
        a pre-defined material color

    """
    from sepal_ui.frontend.styles import TYPES

    if color not in TYPES:
        warnings.warn(
            f'the selected color "{color}" is not a pre-defined material color. It should be one from [{", ".join(TYPES)}]',
            SepalWarning,
        )
        color = TYPES[0]

    return color


@versionadded(version="2.8.0")
def geojson_to_ee(
    geo_json: dict, geodesic: bool = False, encoding: str = "utf-8"
) -> ee.FeatureCollection:
    """
    Transform a geojson object into a featureCollection or a Geometry.

    No sanity check is performed on the initial geo_json. It must respect the
    `__geo_interface__ <https://gist.github.com/sgillies/2217756>`__.

    Args:
        geo_json: a geo_json dictionnary
        geodesic: Whether line segments should be interpreted as spherical geodesics. If false, indicates that line segments should be interpreted as planar lines in the specified CRS. If absent, defaults to True if the CRS is geographic (including the default EPSG:4326), or to False if the CRS is projected. Defaults to False.
        encoding: The encoding of characters. Defaults to "utf-8".

    Returns:
        the created featurecollection
    """
    # from a featureCollection
    if geo_json["type"] == "FeatureCollection":
        for feature in geo_json["features"]:
            if feature["geometry"]["type"] != "Point":
                feature["geometry"]["geodesic"] = geodesic
        features = ee.FeatureCollection(geo_json)
        return features

    # from a single feature
    elif geo_json["type"] == "Feature":
        geom = None
        # Checks whether it is a point
        if geo_json["geometry"]["type"] == "Point":
            coordinates = geo_json["geometry"]["coordinates"]
            longitude = coordinates[0]
            latitude = coordinates[1]
            geom = ee.Geometry.Point(longitude, latitude)
        # for every other geometry simply create a geometry
        else:
            geom = ee.Geometry(geo_json["geometry"], "", geodesic)

        return geom

    # some error handling because we are fancy
    else:
        raise ValueError("Could not convert the geojson to ee.Geometry()")

    return


def check_input(input_: Any, msg: str = ms.utils.check_input.error) -> bool:
    r"""
    Check if the inpupt value is initialized.

    If not raise an error, else return True.

    Args:
        input\_: the input to check
        msg: the message to display if the input is not set

    Return:
        check if the value is initialized
    """
    # by the default the variable is considered valid
    init = True

    # check the collection type that are the only one supporting the len method
    try:
        init = False if len(input_) == 0 else init
    except Exception:
        init = False if input_ is None else init

    if init is False:
        raise ValueError(msg)

    return init


################################################################################
# the soon to be deprecated decorators
#

# fmt: off
catch_errors = deprecated(version='3.0', reason="use sepal_ui.scripts.decorator.catch_errors instead")(sd.catch_errors)
need_ee = deprecated(version='3.0', reason="use sepal_ui.scripts.decorator.need_ee instead")(sd.need_ee)
loading_button = deprecated(version='3.0', reason="use sepal_ui.scripts.decorator.need_ee instead")(sd.loading_button)
switch = deprecated(version='3.0', reason="use sepal_ui.scripts.decorator.switch instead")(sd.switch)
# fmt: on



"""
Model object dedicated to Planet interface.
"""

import asyncio
from datetime import datetime
from typing import Dict, List, Optional, Union

import nest_asyncio
import planet.data_filter as filters
import traitlets as t
from planet import DataClient
from planet.auth import Auth
from planet.exceptions import NoPermission
from planet.http import Session

from sepal_ui.message import ms
from sepal_ui.model import Model

# known problem https://github.com/jupyter/notebook/issues/3397
nest_asyncio.apply()


class PlanetModel(Model):

    SUBS_URL: str = (
        "https://api.planet.com/auth/v1/experimental/public/my/subscriptions"
    )
    "The url of the planet API subscription"

    credentials: List[str] = []
    "list containing [api_key] or pair of [username, password] to log in"

    session: Optional[Session] = None
    "planet.http.session: planet session."

    subscriptions: t.Dict = t.Dict({}).tag(sync=True)
    "All the dictionary info from the available subscriptions"

    active = t.Bool(False).tag(sync=True)
    "Value to determine if at least one subscription has the active true state"

    def __init__(self, credentials: Union[str, List[str]] = "") -> None:
        """
        Planet model helper to connect planet API client and perform requests.

        It can be
        instantiated whether itself or linked with a PlanetView input helper. All the methods
        are aimed to be used without the need of a view.

        Args:
            credentials: planet API key or tuple of username and password of planet explorer.
        """
        self.subscriptions = {}
        self.session = None
        self.active = False

        if credentials:
            self.init_session(credentials)

    def init_session(self, credentials: Union[str, List[str]]) -> None:
        """
        Initialize planet client with api key or credentials. It will handle errors.

        Args:
            credentials: planet API key or username and password pair of planet explorer.
        """
        if not isinstance(credentials, list):
            credentials = [credentials]

        if not all(credentials):
            raise ValueError(ms.planet.exception.empty)

        if len(credentials) == 2:
            self.auth = Auth.from_login(*credentials)
        else:
            self.auth = Auth.from_key(credentials[0])

        self.session = Session(auth=self.auth)
        self._is_active()

        return

    def _is_active(self) -> None:
        """
        check if the key has an associated active subscription and change the state button accordingly.
        """
        # As there is not any key that identify the nicfi contract,
        # let's find though all the subscriptions a representative name
        wildcards = ["Level_0", "Level_1", "Level2"]

        # get the subs from the api key and save them in the model. It will be useful
        # to avoid doing more calls.
        tmp_subscriptions: dict[str, list] = {"nicfi": [], "others": []}
        for sub in self.get_subscriptions():
            for w in wildcards:
                if w in str(sub):
                    tmp_subscriptions["nicfi"].append(sub)
                    break
            if sub not in tmp_subscriptions["nicfi"]:
                tmp_subscriptions["others"].append(sub)

        self.subscriptions = tmp_subscriptions

        states = self.search_status(self.subscriptions)
        self.active = any([next(iter(d.values())) for d in states])

        return

    def get_subscriptions(self) -> dict:
        """
        load the user subscriptions.

        Returns:
            the dictionnary of user subscription or empty list if nothing found
        """
        req = self.session.request("GET", self.SUBS_URL)

        try:
            response = asyncio.run(req)

        except NoPermission:
            raise Exception(
                "You don't have permission to access to this resource. Check your input data."
            )

        except Exception as e:
            raise e

        return response.json() if response.status_code == 200 else {}

    def get_items(
        self,
        aoi: dict,
        start: Union[str, datetime],
        end: Union[str, datetime],
        cloud_cover: float,
        limit_to_x_pages: Optional[int] = None,
    ) -> list:
        """
        Request imagery items from the planet API for the requested dates.

        Args:
            aoi: geojson clipping geometry
            start: the start of the request (YYYY-mm-dd))
            end: the end of the request (YYYY-mm-dd))
            cloud_cover: maximum cloud coverage.
            limit_to_x_pages: number of pages to constrain the search. Defaults to -1 to use all of them.

        Returns:
            items found using the search query

        """
        # cast start and end to str
        start = (
            datetime.strptime(start, "%Y-%m-%d") if isinstance(start, str) else start
        )
        end = datetime.strptime(end, "%Y-%m-%d") if isinstance(end, str) else end

        and_filter = filters.and_filter(
            [
                filters.geometry_filter(aoi),
                filters.range_filter("cloud_cover", lte=cloud_cover),
                filters.date_range_filter("acquired", gt=start),
                filters.date_range_filter("acquired", lt=end),
            ]
        )

        # PSScene3Band and PSScene4Band item type and assets are deprecated
        # since January 2023 so we are now only looking at PSScene
        item_types = ["PSScene"]

        async def _main():
            """
            Create an asyncrhonous function here to avoid making the main get_items as async.

            So we can keep calling get_items without any change.
            """
            client = DataClient(self.session)
            items = await client.search(item_types, and_filter, name="quick_search")
            items.limit = limit_to_x_pages
            items_list = [item async for item in items]
            return items_list

        return asyncio.run(_main())

    @staticmethod
    def search_status(d: dict) -> List[Dict[str, bool]]:
        """
        Get the status of a specific subscription.

        Args:
            d: dictionnary of subscription object

        Returns:
            the (sub.name: status) pairs
        """
        states = []

        for v in d.values():
            for subs in v:
                if "plan" in subs:
                    plan = subs.get("plan")
                    state = True if plan.get("state") == "active" else False
                    states.append({plan.get("name"): state})

        return states



"""
Script to manually activate one of the venv available in the current Jupyter environment.

Application are designed to run on specific venv, created to avoid lib deprecation.
this script allows the user to easily activate one of the venv have already installed in Jupyter and customize it
"""

import argparse
import subprocess
from pathlib import Path

import pandas as pd
from colorama import Fore, init

# init colors for all plateforms
init()

parser = argparse.ArgumentParser(description=__doc__, usage="activate_venv")


def main() -> None:
    """Activate the selected venv."""
    # parse agruments
    parser.parse_args()

    print(
        f"{Fore.CYAN} Welcome to the virtual env activation, loading your venvs... \n{Fore.RESET}"
    )

    # Get usr venvs
    result = subprocess.run(["jupyter", "kernelspec", "list"], stdout=subprocess.PIPE)

    venvs = pd.DataFrame(
        [
            line.split()
            for line in result.stdout.decode("utf-8").splitlines()
            if ".local" not in line
        ],
    ).iloc[1:]

    # Check if user has test envs
    test_venv_path = Path.home() / "module-venv"
    if test_venv_path.exists():
        test_envs = pd.DataFrame(
            list(
                [f"test {el.name}", str(el)]
                for el in test_venv_path.glob("[!.]*")
                if el.is_dir()
            )
        )
        venvs = pd.concat([venvs, test_envs])
    venvs = venvs.reset_index(drop=True)
    venvs.columns = ["env name", "path"]

    print(f"{Fore.CYAN} You can activate any of the following venvs: \n{Fore.RESET}")
    print(venvs)

    valid = False

    while not valid:

        selection = int(
            input(
                f"{Fore.CYAN} Select the venv number you want to activate: \n{Fore.RESET}"
            )
        )

        if selection not in venvs.index.unique():
            print(f"{Fore.RED}Your selection is not valid {Fore.RESET}")
        else:
            valid = True
    # Activate virtual env
    kernel_path = Path(venvs.iloc[selection]["path"])
    test = "module-venv" in str(kernel_path)

    activate_path = "bin/activate" if test else "venv/bin/activate"
    print(f"{Fore.GREEN}Activating: {kernel_path} {Fore.RESET}")

    # Based on https://stackoverflow.com/questions/6943208/activate-a-virtualenv-with-a-python-script
    # It won't return anything to know if the the command was succesfully done
    subprocess.run(
        ["/bin/bash", "--rcfile", str(kernel_path / activate_path)],
    )

    # The following lines won't be executed because the previous subprocess kill the kernel

    # Confirm that we are in the new env
    result = subprocess.run(
        ["echo", "$VIRTUAL_ENV"], shell=True, stdout=subprocess.PIPE
    )

    print(f"The current env is: {result.stdout!r}")

    return


if __name__ == "__main__":
    main()



"""
Customized drawing control.
"""

from copy import deepcopy
from typing import Optional

import geopandas as gpd
import ipyleaflet as ipl
from shapely import geometry as sg

from sepal_ui import color


class DrawControl(ipl.DrawControl):

    m: Optional[ipl.Map] = None
    "the map on which he drawControl is displayed. It will help control the visibility"

    def __init__(self, m: ipl.Map, **kwargs) -> None:
        """
        A custom DrawingControl object to handle edition of features.

        Args:
            m: the map on which he drawControl is displayed
            kwargs: any available arguments from a ipyleaflet.DrawingControl
        """
        # set some default parameters
        options = {"shapeOptions": {"color": color.info}}
        kwargs.setdefault("marker", {})
        kwargs.setdefault("circlemarker", {})
        kwargs.setdefault("polyline", {})
        kwargs.setdefault("rectangle", options)
        kwargs.setdefault("circle", options)
        kwargs.setdefault("polygon", options)

        # save the map in the member of the objects
        self.m = m

        super().__init__(**kwargs)

    def show(self) -> None:
        """
        show the drawing control on the map. and clear it's content.
        """
        self.clear()
        self in self.m.controls or self.m.add(self)

        return

    def hide(self) -> None:
        """
        hide the drawing control from the map, and clear it's content.
        """
        self.clear()
        self not in self.m.controls or self.m.remove(self)

        return

    def to_json(self) -> dict:
        """
        Return the content of the DrawControl data.

        Returned without the styling properties and using a polygonized representation of circles. The output is fully compatible with __geo_interface__.

        Returns:
            the json representation of all the geometries draw on the map
        """
        features = [self.polygonize(feat) for feat in deepcopy(self.data)]
        [feat["properties"].pop("style") for feat in features]

        return {"type": "FeatureCollection", "features": features}

    @staticmethod
    def polygonize(geo_json: dict) -> dict:
        """
        Transform a ipyleaflet circle (a point with a radius) into a GeoJson polygon.

        The methods preserves all the geo_json other attributes.
        If the geometry is not a circle (don't require polygonisation), do nothing.

        Params:
            geo_json: the circle geojson

        Returns:
            the polygonised feature
        """
        if "Point" not in geo_json["geometry"]["type"]:
            return geo_json

        # create shapely point
        center = sg.Point(geo_json["geometry"]["coordinates"])
        point = gpd.GeoSeries([center], crs=4326)

        radius = geo_json["properties"]["style"]["radius"]
        circle = point.to_crs(3857).buffer(radius).to_crs(4326)

        # insert it in the geo_json
        output = geo_json.copy()
        output["geometry"] = circle[0].__geo_interface__

        return output



"""
Model object dedicated to AOI selection.
"""

import json
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

import ee
import geopandas as gpd
import pandas as pd
import traitlets as t
from ipyleaflet import GeoJSON
from typing_extensions import Self

from sepal_ui import color
from sepal_ui.frontend import styles as ss
from sepal_ui.message import ms
from sepal_ui.model import Model
from sepal_ui.scripts import gee
from sepal_ui.scripts import utils as su

__all__ = ["AoiModel"]


class AoiModel(Model):

    # ###########################################################################
    # ###                      dataset const                                  ###
    # ###########################################################################

    FILE: List[Path] = [
        Path(__file__).parents[1] / "data" / "gadm_database.parquet",
        Path(__file__).parents[1] / "data" / "gaul_database.parquet",
    ]
    "Paths to the GADM(0) and GAUL(1) database"

    CODE: List[str] = ["GID_{}", "ADM{}_CODE"]
    "GADM(0) and GAUL(1) administrative codes key format"

    NAME: List[str] = ["NAME_{}", "ADM{}_NAME"]
    "GADM(0) and GAUL(1) naming key format"

    ISO: List[str] = ["GID_0", "ISO 3166-1 alpha-3"]
    "GADM(0) and GAUL(1) iso codes key"

    GADM_BASE_URL: str = (
        "https://geodata.ucdavis.edu/gadm/gadm4.1/json/gadm41_{}_{}.json"
    )
    "The base url to download gadm maps"

    GAUL_ASSET: str = "FAO/GAUL/2015/level{}"
    "The GAUL asset name"

    ASSET_SUFFIX: str = "aoi_"
    "The suffix to identify the asset in GEE"

    # ###########################################################################
    # ###                             const methods                           ###
    # ###########################################################################

    CUSTOM: str = ms.aoi_sel.custom
    "The word displayed for custom method in the relevant lang"

    ADMIN: str = ms.aoi_sel.administrative
    "The word displayed for admin method in the relevant lang"

    METHODS: Dict[str, Dict[str, str]] = {
        "ADMIN0": {"name": ms.aoi_sel.adm[0], "type": ADMIN},
        "ADMIN1": {"name": ms.aoi_sel.adm[1], "type": ADMIN},
        "ADMIN2": {"name": ms.aoi_sel.adm[2], "type": ADMIN},
        "SHAPE": {"name": ms.aoi_sel.vector, "type": CUSTOM},
        "DRAW": {"name": ms.aoi_sel.draw, "type": CUSTOM},
        "POINTS": {"name": ms.aoi_sel.points, "type": CUSTOM},
        "ASSET": {"name": ms.aoi_sel.asset, "type": CUSTOM},
    }
    "The word displayed for all selection methods in the relevant lang"

    # ###########################################################################
    # ###                      widget related traitlets                       ###
    # ###########################################################################

    method: t.Unicode = t.Unicode(None, allow_none=True).tag(sync=True)
    "str: the currently selected method"

    point_json: t.Dict = t.Dict(None, allow_none=True).tag(sync=True)
    "dict: information that will be use to transform the csv into a gdf"

    vector_json: t.Dict = t.Dict(None, allow_none=True).tag(sync=True)
    "dict: information that will be use to transform the vector file into a gdf"

    geo_json: t.Dict = t.Dict(None, allow_none=True).tag(sync=True)
    "dict: the drawn geojson shape"

    admin: t.Unicode = t.Unicode(None, allow_none=True).tag(sync=True)
    "The admin number selected"

    asset_name: t.Unicode = t.Unicode(None, allow_none=True).tag(sync=True)
    "The asset name (only for GEE model)"

    asset_json: t.Dict = t.Dict(None, allow_none=True).tag(sync=True)
    "The asset json description (only for GEE model)"

    name: t.Unicode = t.Unicode(None, allow_none=True).tag(sync=True)
    "The name of the file to create (used only in drawn shaped)"

    # ###########################################################################
    # ###                           model parameters                          ###
    # ###########################################################################

    gee: bool = True
    "either or not the model is bound to gee"

    folder: Union[str, Path] = ""
    "The folder name used in GEE related component, mainly used for debugging"

    default_vector: Optional[Union[str, Path]] = None
    "The default vector file that will be used to produce the gdf. need to be readable by fiona and/or GDAL/OGR"

    default_admin: Optional[str] = None
    "The default administrative area in GADM or GAUL norm"

    default_asset: Optional[str] = None
    "The default asset name, need to point to a readable FeatureCollection"

    # ###########################################################################
    # ###                           model outputs                             ###
    # ###########################################################################

    dst_asset_id: str = ""
    "The exported asset id"

    selected_feature: Optional[Union[ee.Feature, gpd.GeoDataFrame]] = None
    "The Feature associated with a query"

    gdf: Optional[gpd.GeoDataFrame] = None
    "The geodataframe corresponding to the selected AOI"

    feature_collection: Optional[ee.FeatureCollection] = None
    "The feature Collection generated by the parameters (only for GEE models)"

    ipygeojson: Optional[GeoJSON] = None
    "The representation of the AOI as a ipyleaflet layer"

    def __init__(
        self,
        gee: bool = True,
        vector: Optional[Union[str, Path]] = None,
        asset: Optional[Union[str, Path]] = None,
        admin: Optional[str] = None,
        folder: Union[str, Path] = "",
    ) -> None:
        """
        An Model object dedicated to the sorage and the manipulation of aoi.

        It is meant to be used with the AoiView object (embeded in the AoiTile).
        By using this you will be able to provide your application with aoi as an ee_object
        or a gdf, depending if you activated the ee binding or not.
        The class also provide insight on your aoi geometry.

        Args:
            gee: wether or not the aoi selector should be using the EarthEngine binding
            vector: the path to the default vector object
            admin: the administrative code of the default selection. Need to be GADM if ee==False and GAUL 2015 if ee==True.
            asset: the default asset. Can only work if ee==True
            folder: the init GEE asset folder where the asset selector should start looking (debugging purpose)

        .. deprecated:: 2.3.2
            'asset_name' will be used as variable to store 'ASSET' method info. To get the destination saved asset id, please use 'dst_asset_id' variable.

        """
        super().__init__()

        # the ee retated informations
        self.gee = gee
        if gee:
            su.init_ee()
            self.folder = str(folder) or ee.data.getAssetRoots()[0]["id"]

        # set default values
        self.set_default(vector, admin, asset)

    def set_default(
        self,
        vector: Optional[Union[str, Path]] = None,
        admin: Optional[str] = None,
        asset: Optional[Union[str, Path]] = None,
    ) -> Self:
        """
        Set the default value of the object and create a gdf/feature_collection out of it.

        Args:
            vector: the default vector file that will be used to produce the gdf. need to be readable by fiona and/or GDAL/OGR
            admin: the default administrative area in GADM or GAUL norm
            asset: the default asset name, need to point to a readable FeatureCollection
        """
        # save the default values
        self.default_vector = vector
        self.default_asset = self.asset_name = str(asset) if asset else None
        self.asset_json = (
            {"pathname": asset, "column": "ALL", "value": None} if asset else None
        )
        self.default_admin = self.admin = admin

        # cast the vector to json
        self.vector_json = (
            {"pathname": str(vector), "column": "ALL", "value": None}
            if vector
            else None
        )

        # cast the asset to json
        self.asset_json = (
            {"pathname": asset, "column": "ALL", "value": None} if asset else None
        )

        # set the default gdf if possible
        if self.vector_json is not None:
            self.set_object("SHAPE")
        elif self.admin:
            self.set_object("ADMIN0")  # any level will work
        elif self.asset_json is not None:
            self.set_object("ASSET")

        return self

    def set_object(self, method: str = "") -> Self:
        """
        set the object (gdf/featurecollection) based on the model inputs.

        The method can be manually overwritten by setting the ``method`` parameter.

        Args:
            method: a model loading method
        """
        # clear the model output if existing
        self.clear_output()

        # overwrite self.method
        self.method = method or self.method

        if self.method in ["ADMIN0", "ADMIN1", "ADMIN2"]:
            self._from_admin(self.admin)
        elif self.method == "POINTS":
            self._from_points(self.point_json)
        elif self.method == "SHAPE":
            self._from_vector(self.vector_json)
        elif self.method == "DRAW":
            self._from_geo_json(self.geo_json)
        elif self.method == "ASSET":
            self._from_asset(self.asset_json)
        else:
            raise Exception(ms.aoi_sel.exception.no_inputs)

        return self

    def _from_asset(self, asset_json: dict) -> Self:
        """Set the ee.FeatureCollection output from an existing asset."""
        if not (asset_json["pathname"]):
            raise Exception(ms.aoi_sel.exception.no_asset)

        if asset_json["column"] != "ALL":
            if asset_json["value"] is None:
                raise Exception(ms.aoi_sel.exception.no_value)

        # set the name
        self.name = Path(asset_json["pathname"]).stem.replace(self.ASSET_SUFFIX, "")
        self.asset_name = asset_json["pathname"]
        ee_col = ee.FeatureCollection(asset_json["pathname"])

        if asset_json["column"] != "ALL":

            column = asset_json["column"]
            value = asset_json["value"]
            ee_col = ee_col.filterMetadata(column, "equals", value)
            self.name = f"{self.name}_{column}_{value}"

        # set the feature collection
        self.feature_collection = ee_col

        # create a gdf form te feature_collection
        features = self.feature_collection.getInfo()["features"]
        self.gdf = gpd.GeoDataFrame.from_features(features).set_crs(epsg=4326)

        return self

    def _from_points(self, point_json: dict) -> Self:
        """
        Set the object output from a csv json.

        Args:
            point_json: the geo_interface description of the points
        """
        if not all(point_json.values()):
            raise Exception(ms.aoi_sel.exception.uncomplete)

        # cast the pathname to pathlib Path
        point_file = Path(point_json["pathname"])

        # check that the columns are well set
        values = [v for v in point_json.values()]
        if not len(values) == len(set(values)):
            raise Exception(ms.aoi_sel.exception.duplicate_key)

        # create the gdf
        df = pd.read_csv(point_file, sep=None, engine="python")
        self.gdf = gpd.GeoDataFrame(
            df,
            crs="EPSG:4326",
            geometry=gpd.points_from_xy(
                df[point_json["lng_column"]], df[point_json["lat_column"]]
            ),
        )

        # set the name
        self.name = point_file.stem

        if self.gee:
            # transform the gdf to ee.FeatureCollection
            self.feature_collection = ee.FeatureCollection(self.gdf.__geo_interface__)

            # export as a GEE asset
            self.export_to_asset()

        return self

    def _from_vector(self, vector_json: dict) -> Self:
        """
        Set the object output from a vector json.

        Args:
            vector_json: the dict describing the vector file, and column filter
        """
        if not (vector_json["pathname"]):
            raise Exception(ms.aoi_sel.exception.no_file)

        if vector_json["column"] != "ALL":
            if vector_json["value"] is None:
                raise Exception(ms.aoi_sel.exception.no_value)

        # cast the pathname to pathlib Path
        vector_file = Path(vector_json["pathname"])

        # create the gdf
        self.gdf = gpd.read_file(vector_file).to_crs("EPSG:4326")

        # set the name using the file stem
        self.name = vector_file.stem

        # filter it if necessary
        if vector_json["value"] is not None:
            self.gdf = self.gdf[self.gdf[vector_json["column"]] == vector_json["value"]]
            self.name = f"{self.name}_{vector_json['column']}_{vector_json['value']}"

        if self.gee:
            # transform the gdf to ee.FeatureCollection
            self.feature_collection = su.geojson_to_ee(self.gdf.__geo_interface__)

            # export as a GEE asset
            self.export_to_asset()

        return self

    def _from_geo_json(self, geo_json: dict) -> Self:
        """
        Set the gdf output from a geo_json.

        Args:
            geo_json: the __geo_interface__ dict of a geometry drawn on the map
        """
        if not geo_json:
            raise Exception(ms.aoi_sel.exception.no_draw)

        # remove the style property from geojson as it's not recognize by geopandas and gee
        for feat in geo_json["features"]:
            if "style" in feat["properties"]:
                del feat["properties"]["style"]

        # create the gdf
        self.gdf = gpd.GeoDataFrame.from_features(geo_json).set_crs(epsg=4326)

        # normalize the name
        self.name = su.normalize_str(self.name)

        if self.gee:
            # transform the gdf to ee.FeatureCollection
            self.feature_collection = su.geojson_to_ee(self.gdf.__geo_interface__)

            # export as a GEE asset
            self.export_to_asset()
        else:
            # save the geojson in downloads
            path = Path("~", "downloads", "aoi").expanduser()
            path.mkdir(
                exist_ok=True, parents=True
            )  # if nothing have been run the downloads folder doesn't exist
            self.gdf.to_file(path / f"{self.name}.geojson", driver="GeoJSON")

        return self

    def _from_admin(self, admin: str) -> Self:
        """
        Set the object according to given an administrative number in the GADM norm.

        Args:
            admin: the admin code corresponding to FAO GAUl (if gee) or GADM
        """
        if not admin:
            raise Exception(ms.aoi_sel.exception.no_admlyr)

        # get the admin level corresponding to the given admin code
        df = pd.read_parquet(self.FILE[self.gee]).astype(str)

        # extract the first element that include this administrative code and set the level accordingly
        is_in = df.filter([self.CODE[self.gee].format(i) for i in range(3)]).isin(
            [admin]
        )

        if not is_in.any().any():
            raise Exception(ms.aoi_sel.exception.invalid_code)

        index = 3 if self.gee else -1
        level = is_in[~((~is_in).all(axis=1))].idxmax(1).iloc[0][index]

        if self.gee:

            # get the feature_collection
            self.feature_collection = ee.FeatureCollection(
                self.GAUL_ASSET.format(level)
            ).filter(ee.Filter.eq(f"ADM{level}_CODE", int(admin)))

            # transform it into gdf
            features = self.feature_collection.getInfo()["features"]
            self.gdf = gpd.GeoDataFrame.from_features(features).set_crs(epsg=4326)

        else:
            # save the country iso_code
            iso_3 = admin[:3]

            # read the data from server
            level_gdf = gpd.read_file(self.GADM_BASE_URL.format(iso_3, level))
            level_gdf.rename(columns={"COUNTRY": "NAME_0"}, inplace=True)
            self.gdf = level_gdf[level_gdf[self.CODE[self.gee].format(level)] == admin]

        # set the name using the layer
        r = df[df[self.CODE[self.gee].format(level)] == admin].iloc[0]
        names = [
            su.normalize_str(r[self.NAME[self.gee].format(i)])
            if i
            else r[self.ISO[self.gee]]
            for i in range(int(level) + 1)
        ]
        self.name = "_".join(names)

        return self

    def clear_output(self) -> Self:
        """
        Clear the output of the aoi selector without changing the traits and/or the parameters.
        """
        # reset the outputs
        self.gdf = None
        self.feature_collection = None
        self.ipygeojson = None
        self.selected_feature = None
        self.dst_asset_id = None

        return self

    def clear_attributes(self) -> Self:
        """
        Return all attributes to their default state.

        Note:
            Set the default setting as current object.
        """
        # keep the default
        admin = self.default_admin
        vector = self.default_vector
        asset = self.default_asset

        # delete all the traits
        [setattr(self, attr, None) for attr in self.trait_names()]

        # reset the outputs
        self.clear_output()

        # reset the default
        self.set_default(vector, admin, asset)

        return self

    def get_columns(self) -> List[str]:
        """
        Retrieve the columns or variables from self excluding geometries and gee index.

        Returns:
            sorted list of column names
        """
        if self.gdf is None:
            raise Exception(ms.aoi_sel.exception.no_gdf)

        if self.gee:
            aoi_ee = ee.Feature(self.feature_collection.first())
            columns = aoi_ee.propertyNames().getInfo()
            list_ = [
                col for col in columns if col not in ["system:index", "Shape_Area"]
            ]
        else:
            list_ = list(set(["geometry"]) ^ set(self.gdf.columns.to_list()))

        return sorted(list_)

    def get_fields(self, column: str) -> List[str]:
        """
        Retrieve the fields from a column.

        Args:
            A column name to query over the asset

        Returns:
            sorted list of fields value

        """
        if self.gdf is None:
            raise Exception(ms.aoi_sel.exception.no_gdf)

        if self.gee:
            fields = self.feature_collection.distinct(column).aggregate_array(column)
            list_ = fields.getInfo()
        else:
            list_ = self.gdf[column].to_list()

        return sorted(list_)

    def get_selected(
        self, column: str, field: str
    ) -> Union[ee.Feature, gpd.GeoDataFrame]:
        """
        Select an ee object based on selected column and field.

        Args:
            column: the selected column in the dataset
            field: the value to search in the selected column

        Returns:
            The Feature associated with the query
        """
        if self.gdf is None:
            raise Exception(ms.aoi_sel.exception.no_gdf)

        if self.gee:
            selected_feature = self.feature_collection.filterMetadata(
                column, "equals", field
            )
        else:
            selected_feature = self.gdf[self.gdf[column] == field]

        return selected_feature

    def total_bounds(self) -> Tuple[float, float, float, float]:
        """
        Reproduce the behaviour of the total_bounds method from geopandas.

        Returns:
            minxx, miny, maxx, maxy
        """
        if self.gdf is None:
            raise ValueError(ms.aoi_sel.exception.no_gdf)

        return self.gdf.total_bounds.tolist()

    def export_to_asset(self) -> Self:
        """
        Export the feature_collection as an asset (only for ee model).
        """
        asset_name = self.ASSET_SUFFIX + self.name
        asset_id = str(Path(self.folder, asset_name))

        self.dst_asset_id = asset_id

        # check if the table already exist
        if asset_id in [a["name"] for a in gee.get_assets(self.folder)]:
            return self

        # check if the task is running
        if gee.is_running(asset_name):
            return self

        # run the task
        task_config = {
            "collection": self.feature_collection,
            "description": asset_name,
            "assetId": asset_id,
        }

        task = ee.batch.Export.table.toAsset(**task_config)
        task.start()

        return self

    def get_ipygeojson(self, style: Optional[dict] = None) -> GeoJSON:
        """
        Converts current geopandas object into ipyleaflet GeoJSON.

        Args:
            style: the predifined style of the aoi. It's by default using a "success" ``sepal_ui.color`` with 0.5 transparent fill color. It can be completly replace by a fully qualified `style dictionnary <https://ipyleaflet.readthedocs.io/en/latest/layers/geo_json.html>`__. Use the ``sepal_ui.color`` object to define any color to remain compatible with light and dark theme.

        Returns:
            The geojson layer of the aoi gdf, ready to use in a Map
        """
        if self.gdf is None:
            raise Exception(ms.aoi_sel.exception.no_gdf)

        # read the data from geojson and add the name as a property of the shape
        # useful when handler are added from ipyleaflet
        data = json.loads(self.gdf.to_json())
        for f in data["features"]:
            f["properties"]["name"] = self.name

        # adapt the style to the theme
        if style is None:
            style = json.loads((ss.JSON_DIR / "aoi.json").read_text())
            style.update(color=color.success, fillColor=color.success)

        # create a GeoJSON object
        # attribution="SEPAL(c)" is not recognized yet
        # https://github.com/jupyter-widgets/ipyleaflet/issues/847
        self.ipygeojson = GeoJSON(data=data, style=style, name="aoi")

        return self.ipygeojson



"""
All directories used in the application. 
"""

from pathlib import Path

# this directory is the root directory of all sepal dashboard app.
# please make sure that any result that you produce is embeded inside this folder
# create a folder adapted to your need inside this folder to save anything in sepal
module_dir = Path.home() / "module_results"
module_dir.mkdir(exist_ok=True)



"""
Initialization of the ``Translator`` used in the sepal-ui.

Can be accessed as: 

.. code-block::
    
    from sepal_ui.message import ms
"""

from pathlib import Path

from sepal_ui.translator import Translator

ms = Translator(Path(__file__).parent)



"""
The translator object allow developer to suport translation for their application.
"""

import json
from configparser import ConfigParser
from pathlib import Path
from typing import List, Tuple, Union

import pandas as pd
from box import Box
from deprecated.sphinx import deprecated, versionadded

from sepal_ui.conf import config_file


class Translator(Box):

    _protected_keys = [
        "find_target",
        "search_key",
        "sanitize",
        "_update",
        "missing_keys",
        "available_locales",
        "merge_dict",
        "delete_empty",
    ] + dir(Box)
    "keys that cannot be used as var names as they are protected for methods"

    def __init__(
        self, json_folder: Union[str, Path], target: str = "", default: str = "en"
    ) -> None:
        """
        Python ``Box`` of ``Box`` representing all the nested translation key, value pairs.

        It reads 2 Json files, the first one being the source language (usually English) and the second one the target language.
        It will replace in the source dictionary every key that exist in both json dictionaries. Following this procedure, every message that is not translated can still be accessed in the source language.
        To access the dictionary keys, instead of using [], you can simply use key name as in an object ex: translator.first_key.secondary_key.
        There are no depth limits, just respect the snake_case convention when naming your keys in the .json files.
        5 internal keys are created upon initialization (there name cannot be used as keys in the translation message):

        -   (str) _default : the default locale of the translator
        -   (str) _targeted : the initially requested language. Use to display debug information to the user agent
        -   (str) _target : the target locale of the translator
        -   (bool) _match : if the target language match the one requested one by user, used to trigger information in appBar
        -   (str) _folder : the path to the l10n folder

        Args:
            json_folder: The folder where the dictionaries are stored
            target: The language code (IETF BCP 47) of the target lang (it should be the same as the target dictionary). Default to either the language specified in the parameter file or the default one.
            default: The language code (IETF BCP 47) of the source lang. default to "en" (it should be the same as the source dictionary)
        """
        # the name of the 5 variables that cannot be used as init keys
        FORBIDDEN_KEYS = ["_folder", "_default", "_target", "_targeted", "_match"]

        # init the box with the folder
        folder = Path(json_folder)

        # reading the default dict
        default_dict = self.merge_dict(folder / default)

        # create a dictionary in the target language
        targeted, target = self.find_target(folder, target)
        target = target or default
        target_dict = self.merge_dict(folder / target)

        # evaluate the matching of requested and obtained values
        match = targeted == target

        # create the composite dictionary
        ms_dict = self._update(default_dict, target_dict)

        # check if forbidden keys are being used
        # this will raise an error if any
        [self.search_key(ms_dict, k) for k in FORBIDDEN_KEYS + self._protected_keys]

        # # unpack the json as a simple namespace
        ms_json = json.dumps(ms_dict)
        ms_boxes = json.loads(ms_json, object_hook=lambda d: Box(**d, frozen_box=True))

        private_keys = {
            "_folder": str(folder),
            "_default": default,
            "_targeted": targeted,
            "_target": target,
            "_match": match,
        }

        # the final box is not frozen
        # waiting for an answer here: https://github.com/cdgriffith/Box/issues/223
        # it the meantime it's easy to call the translator using a frozen_box argument
        super(Box, self).__init__(**private_keys, **ms_boxes)

    @versionadded(version="2.7.0")
    @staticmethod
    def find_target(folder: Path, target: str = "") -> Tuple[str, str]:
        """
        find the target language in the available language folder.

        given a folder and a target lang, this function returns the closest language available in the folder
        If nothing is found falling back to any working subvariety and return None if it doesn't exist

        Args:
            folder: the folder where the languages dictionnaries are stored
            target: the target lang in IETF BCP 47. If not specified, the value in the sepal-ui config file will be used

        Returns:
            the targeted language code, the closest lang in IETF BCP 47
        """
        # init lang
        lang = ""

        # if target is not set try to find one in the config file
        # exit with none if the config file is not yet existing
        if target == "":
            if config_file.is_file():
                config = ConfigParser()
                config.read(config_file)
                target = config.get("sepal-ui", "locale", fallback="en")
            else:
                return ("", "en")

        # first scenario the target is available
        if (folder / target).is_dir():
            lang = target
        # second senario the "main lang" is set
        elif (folder / target[:2]).is_dir():
            lang = target[:2]
        # third scenario we search for any closely related language
        else:
            try:
                f = next(f for f in folder.glob(f"{target[:2]}-*") if f.is_dir())
                lang = f.stem
            except StopIteration:
                pass

        return (target, lang)

    def search_key(self, d: dict, key: str) -> None:
        """
        Search a specific key in the d dictionary and raise an error if found.

        Args:
            d: the dictionary to study
            key: the key to look for
        """
        if key in d:
            msg = f"You cannot use the key {key} in your translation dictionary"
            raise Exception(msg)

        for k, v in d.items():
            if isinstance(v, dict):
                return self.search_key(v, key)

    @classmethod
    def sanitize(cls, d: Union[dict, list]) -> dict:
        """
        Identify numbered dictionnaries embeded in the dict and transform them into lists.

        This function is an helper to prevent deprecation after the introduction of pontoon for translation.
        The user is now force to use keys even for numbered lists. SimpleNamespace doesn't support integer indexing
        so this function will transform back this "numbered" dictionnary (with integer keys) into lists.

        Args:
            d: the dictionnary to sanitize

        Returns:
            the sanitized dictionnary
        """
        ms = d.copy()

        # create generator based on input type
        if isinstance(ms, dict):
            gen = ms.items()
        elif isinstance(ms, list):
            gen = enumerate(ms)

        # loop into the keys of the dict modify them
        for k, v in gen:
            if isinstance(v, dict):
                tmp = v
                if len(tmp) and all([k.isnumeric() for k in tmp]):
                    tmp = list(tmp.values())
                ms[k] = cls.sanitize(tmp)
            else:
                ms[k] = v

        return ms

    def _update(self, d: dict, u: dict) -> dict:
        """
        Update the fallback dictionnaire (d) values with the keys that exist in the target (u) dictionnaire.

        Args:
            d: The fallback dictionary
            u: the target dctionnary

        Returns:
            The updated dictionnay
        """
        ms = d.copy()

        for k, v in d.items():
            if isinstance(v, dict):
                ms[k] = self._update(v, u.get(k, {}))
            else:
                ms[k] = u.get(k, v)

        return ms

    @deprecated(version="2.9.0", reason="Not needed with automatic translators")
    def missing_keys(self):
        """Nothing."""
        pass

    def available_locales(self) -> List[str]:
        """
        Return the available locales in the l10n folder.

        Returns:
            the list of str codes
        """
        return [f.name for f in Path(self._folder).glob("[!^._]*") if f.is_dir()]

    @versionadded(version="2.7.0")
    @classmethod
    def merge_dict(cls, folder: Path) -> dict:
        """
        Gather all the .json file in the provided l10n folder as 1 single json dict.

        The json dict will be sanitysed and the key will be used as if they were coming from 1 single file.
        be careful with duplication. empty string keys will be removed.

        Args:
            folder: the folder where all the .json files are stored

        Returns:
            the json dict with all the keys

        """
        final_json = {}
        for f in folder.glob("*.json"):
            tmp_dict = cls.delete_empty(json.loads(f.read_text()))
            final_json = {**final_json, **cls.sanitize(tmp_dict)}

        return final_json

    @versionadded(version="2.8.1")
    @classmethod
    def delete_empty(cls, d: dict) -> dict:
        """
        Remove empty strings ("") recursively from the dictionaries.

        This is to prevent untranslated strings from Crowdin to be uploaded. The dictionnary must only embed dictionnaries and no lists.

        Args:
            d: the dictionnary to sanitize

        Returns:
            the sanitized dictionnary

        """
        for k, v in list(d.items()):
            if isinstance(v, dict):
                cls.delete_empty(v)
            elif v == "":
                d.pop(k)

        return d

    @versionadded(version="2.10.0")
    def key_use(self, folder: Path, name: str) -> List[str]:
        """
        Parse all the files in the folder and check if keys are all used at least once.

        Return the unused key names.

        .. warning::

            Don't forget that there are many ways of calling Translator variables
            (getattr, save.cm.xxx in another variable etc...) SO don't forget to check
            manually the variables suggested by this method before deleting them

        Args:
            folder: The application folder using this translator data
            name: the name use by the translator in this app (usually "cm")

        Returns:
            the list of unused keys
        """
        # cannot set FORBIDDEN_KEY in the Box as it would lock another key
        FORBIDDEN_KEYS = ["_folder", "_default", "_target", "_targeted", "_match"]

        # sanitize folder
        folder = Path(folder)

        # get all the python files recursively
        py_files = []
        all_files = [f for f in folder.glob("**/*") if f.suffix in [".py", ".ipynb"]]
        for f in all_files:
            generated_files = [".ipynb_checkpoints", "__pycache__"]
            if all([err not in str(f) for err in generated_files]):
                py_files.append(f)

        # get the flat version of all keys
        keys = list(set(pd.json_normalize(self).columns) ^ set(FORBIDDEN_KEYS))

        # init the unused keys list
        unused_keys = []

        for k in keys:

            # by default we consider that the is never used
            is_present = False

            # read each python file and search for the pattern of the key
            # if it's find change status of the counter and exit the search
            for f in py_files:
                tmp = f.read_text()
                if f"{name}.{k}" in tmp:
                    is_present = True
                    break

            # if nothing is find, the value is still False and the key can be
            # added to the list
            is_present or unused_keys.append(k)

        return unused_keys



#!/usr/bin/python3

"""
Script to create an automatic overwrite of all the object in ipyvuetify. We want to add SepalWidgets to the list of parent class.

This script should be run only by maintainer when changes are made to ipyvuetify itself. Please report to the issue tracker if any class is not available.
"""

import argparse
from datetime import datetime
from pathlib import Path

import ipyvuetify as v

parser = argparse.ArgumentParser(description=__doc__, usage="sepal_ipyvuetify")


def disclaimer() -> str:
    """Return the module docstring."""
    return (
        '"""\n'
        "All the ipyvuetify class override as SepalWidgets\n"
        "\n"
        "This file overwrite all the widgets generating by ipyvuetify to add the SepalWidget class as a parent.\n"
        "It should not be modified from here as it's automatically generated.\n"
        "\n"
        f"last update: {datetime.now():%Y-%m-%d}\n"
        '"""\n'
        "\n"
    )


def imports() -> str:
    """Return the import statements."""
    return (
        "import ipyvuetify as v\n"
        "from sepal_ui.sepalwidgets.sepalwidget import SepalWidget\n"
        "\n"
    )


def klass(klass: str) -> str:
    """Return the class line."""
    return f"class {klass}(v.{klass}, SepalWidget):\n" "    pass\n" "\n"


def is_widget(klass: str) -> bool:
    """Return True i the class is a widget."""
    not_hidden = not klass.startswith("__")
    not_main = klass != "VuetifyWidget"

    return not_hidden and not_main


def main() -> None:
    """Launch the process."""
    # parse agruments
    parser.parse_args()

    # name of the created file
    filename = Path(__file__).parents[1] / "sepalwidgets" / "sepal_ipyvuetify.py"

    # write the content of the file
    with filename.open("w") as f:
        f.write(disclaimer())
        f.write(imports())
        f.write(klass("Html"))  # not included in the c_list

        c_list = [c for c in dir(v.generated) if is_widget(c)]
        for c in c_list:
            f.write(klass(c))

    return


if __name__ == "__main__":
    main()
